"""lg-rez / features / Tâches planifiées

Planification, liste, annulation, exécution de tâches planifiées

"""

import datetime

from discord.ext import commands

from lgrez import config
from lgrez.blocs import env, gsheets, tools
from lgrez.bdd import (Joueur, Action, Role, Camp, Utilisation, Ciblage,
                       CandidHaro, CandidHaroType, UtilEtat, CibleType, Vote)
from lgrez.features import gestion_actions


def export_vote(vote, utilisation):
    """Enregistre un vote/les actions résolues dans le GSheet ad hoc.

    Écrit dans le GSheet ``LGREZ_DATA_SHEET_ID``. Peut être écrasé
    pour une autre implémentation.

    Args:
        vote (.bdd.Vote): le vote concerné, ou ``None`` pour une action.
        utilisation (.bdd.Utilisation): l'utilisation qui vient d'être
            effectuée. Doit être remplie (:attr:`.bdd.Utilisation.is_filled`).

    Raises:
        RuntimeError: si la variable d'environnement ``LGREZ_DATA_SHEET_ID``
            n'est pas définie.
    """
    if vote and not isinstance(vote, Vote):
        vote = Vote[vote]       # str -> Vote

    joueur = utilisation.action.joueur
    if vote == Vote.cond:
        sheet_name = config.db_votecond_sheet
        data = [joueur.nom, utilisation.cible.nom]
    elif vote == Vote.maire:
        sheet_name = config.db_votemaire_sheet
        data = [joueur.nom, utilisation.cible.nom]
    elif vote == Vote.loups:
        sheet_name = config.db_voteloups_sheet
        data = [joueur.nom, joueur.camp.slug, utilisation.cible.nom]
    else:
        sheet_name = config.db_actions_sheet
        recap = "\n+\n".join(
            f"{action.base.slug}({last_util.decision})"
            for action in joueur.actions_actives
            if ((last_util := action.derniere_utilisation)
                and last_util.is_filled         # action effectuée
                and last_util.ts_decision.date() == datetime.date.today())
                # Et dernière décision aujourd'hui ==> on met dans le TDB
        )
        data = [joueur.nom, joueur.role.slug, joueur.camp.slug, recap]

    LGREZ_DATA_SHEET_ID = env.load("LGREZ_DATA_SHEET_ID")
    sheet = gsheets.connect(LGREZ_DATA_SHEET_ID).worksheet(sheet_name)
    timestamp = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
    sheet.append_row([timestamp, *data], value_input_option="USER_ENTERED")


async def get_cible(ctx, action, base_ciblage, first=None):
    """Demande une cible à l'utilisateur.

    Args:
        ctx (~discord.ext.commands.Context): le contexte de commande.
        action: (.bdd.Action): action pour laquelle on cherche une cible.
        base_ciblage (.bdd.BaseCiblage): ciblage à demander.
        first (str): proposition initale du joueur (passée comme argument
            d'une commande).

    Returns:
        Union[.bdd.Joueur, .bdd.Role, .bdd.Camp, bool, str]: La
        cible sélectionnée, selon le type de ciblage.

    Réalise les interactions adéquates en fonction du type du base_ciblage,
    vérifie le changement de cible le cas échéant.
    """
    phrase = base_ciblage.phrase.rstrip()

    # ou_vide = ("" if base_ciblage.obligatoire
    #            else ", ou :no_entry_sign: pour laisser vide")
    if not base_ciblage.obligatoire:
        await ctx.send("[Étape non obligatoire, caractère non pris en "
                       "compte (pour l'instant) par le bot, dommage]")

    stop = f"({tools.code(config.stop_keywords[0])} pour annuler)"

    if base_ciblage.type == CibleType.joueur:
        res = await tools.boucle_query_joueur(
            ctx, cible=first, message=(f"{phrase}\n\n*Écris simplement le "
                                       f"nom du joueur ci-dessous {stop} :*")
        )
    elif base_ciblage.type == CibleType.vivant:
        res = await tools.boucle_query_joueur(
            ctx, cible=first, filtre=Joueur.est_vivant,
            message=(f"{phrase}\n\n*Écris simplement le nom du joueur "
                     f"(vivant) ci-dessous {stop} :*")
        )
    elif base_ciblage.type == CibleType.mort:
        res = await tools.boucle_query_joueur(
            ctx, cible=first, filtre=Joueur.est_mort,
            message=(f"{phrase}\n\n*Écris simplement le nom du mort "
                     f"ci-dessous {stop} :*")
        )
    elif base_ciblage.type == CibleType.role:
        res = await tools.boucle_query(
            ctx, Role, Role.nom, cible=first,
            message=(f"{phrase}\n\n*Écris simplement le nom du rôle "
                     f"ci-dessous {stop} :*")
        )
    elif base_ciblage.type == CibleType.camp:
        res = await tools.boucle_query(
            ctx, Camp, Camp.nom, cible=first,
            message=(f"{phrase}\n\n*Écris simplement le nom du camp "
                     f"ci-dessous {stop} :*")
        )
    elif base_ciblage.type == CibleType.booleen:
        message = await ctx.send(f"{phrase}\n\n*{stop}*")
        if first:
            await ctx.send(quote_bloc(first))
        res = await tools.yes_no(message, first_text=first)
    elif base_ciblage.type == CibleType.texte:
        if first:
            res = first
        else:
            await ctx.send(
                f"{phrase}\n\n*Réponds simplement ci-dessous {stop} :*"
            )
            mess = await tools.wait_for_message_here(ctx)
            res = mess.content

    if base_ciblage.doit_changer:
        derniere_util = action.utilisations.filter(
            ~Utilisation.is_open).order_by(Utilisation.ts_close.desc()).first()
        if derniere_util and derniere_util.etat == UtilEtat.validee:
            # Dernière utilisation validée : comparaison avec ciblages
            # de même prio que le ciblage en cours de demande
            cibles = [cib.valeur for cib in derniere_util.ciblages
                      if cib.base.prio == base_ciblage.prio]

            if res in cibles:       # interdit !
                await ctx.send(
                    f":stop_sign: {res} déjà ciblé(e) lors de la "
                    "précédente utilisation, merci de changer :stop_sign:\n"
                    "*(`@MJ` si contestation)*\n——————————"
                )
                await tools.sleep(ctx, 2)
                # On re-demande (ptite réccurence)
                res = await get_cible(ctx, action, base_ciblage)

    return res


class _BaseCiblageForVote():
    """Mock un objet BaseCiblage pour représenter la cible d'un vote"""
    def __init__(self, phrase):
        self._id = 0
        self.base_action = None
        self.slug = "cible"
        self.type = CibleType.vivant
        self.prio = 1
        self.phrase = phrase
        self.obligatoire = True     # vote blanc non pris en compte
        self.doit_changer = False


class VoterAgir(commands.Cog):
    """Commandes de vote et d'action de rôle"""

    @commands.command()
    @tools.vivants_only
    @tools.private
    async def vote(self, ctx, *, cible=None):
        """Vote pour le condamné du jour

        Args:
            cible: nom du joueur contre qui tu veux diriger ton vote.

        Cette commande n'est utilisable que lorsqu'un vote pour le
        condamné est en cours, pour les joueurs ayant le droit de voter.

        Le bot t'enverra un message à l'ouverture de chaque vote.

        La commande peut être utilisée autant que voulu pour changer
        de cible tant que le vote est en cours.
        """
        joueur = Joueur.from_member(ctx.author)
        try:
            vaction = joueur.action_vote(Vote.cond)
        except RuntimeError:
            await ctx.send("Minute papillon, le jeu n'est pas encore lancé !")
            return

        # Vérification vote en cours
        if not joueur.votant_village:
            await ctx.send("Tu n'as pas le droit de participer à ce vote.")
            return
        if not vaction.is_open:
            await ctx.send("Pas de vote pour le condamné de jour en cours !")
            return

        util = vaction.derniere_utilisation

        # Choix de la cible
        haros = CandidHaro.query.filter_by(type=CandidHaroType.haro).all()
        harotes = [haro.joueur.nom for haro in haros]
        pseudo_bc = _BaseCiblageForVote(
            "Contre qui veux-tu voter ? (vote actuel : "
            f"{tools.bold(util.cible.nom if util.cible else 'aucun')})\n"
            f"(harotés : {', '.join(harotes) or 'aucun :pensive:'})"
        )
        cible = await get_cible(ctx, vaction, pseudo_bc, cible)

        # Test si la cible est sous le coup d'un haro
        cible_ds_haro = CandidHaro.query.filter_by(
            joueur=cible, type=CandidHaroType.haro).all()
        if not cible_ds_haro:
            mess = await ctx.send(
                f"{cible.nom} n'a pas (encore) subi ou posté de haro ! "
                "Si c'est toujours le cas à la fin du vote, ton vote sera "
                "compté comme blanc... \n Veux-tu continuer ?"
            )
            if not await tools.yes_no(mess):
                await ctx.send("Compris, mission aborted.")
                return

        if not vaction.is_open:
            # On revérifie, si ça a fermé entre temps !!
            await ctx.send("Le vote pour le condamné du jour a fermé "
                           "entre temps, pas de chance !")
            return

        # Modification en base
        if util.ciblages:       # ancien ciblage
            Ciblage.delete(*util.ciblages)
        Ciblage(utilisation=util, joueur=cible).add()
        util.ts_decision = datetime.datetime.now()
        util.etat = UtilEtat.remplie
        util.update()

        async with ctx.typing():
            # Écriture dans sheet Données brutes
            export_vote(Vote.cond, util)

        await ctx.send(
            f"Vote contre {tools.bold(cible.nom)} bien pris en compte.\n"
            + tools.ital("Tu peux modifier ton vote autant que nécessaire "
                         "avant sa fermeture.")
        )


    @commands.command()
    @tools.vivants_only
    @tools.private
    async def votemaire(self, ctx, *, cible=None):
        """Vote pour le nouveau maire

        Args:
            cible: nom du joueur pour lequel tu souhaites voter.

        Cette commande n'est utilisable que lorsqu'une élection pour le
        maire est en cours, pour les joueurs ayant le droit de voter.

        Le bot t'enverra un message à l'ouverture de chaque vote.

        La commande peut être utilisée autant que voulu pour changer de
        cible tant que le vote est en cours.
        """
        joueur = Joueur.from_member(ctx.author)
        try:
            vaction = joueur.action_vote(Vote.maire)
        except RuntimeError:
            await ctx.send("Minute papillon, le jeu n'est pas encore lancé !")
            return

        # Vérification vote en cours
        if not joueur.votant_village:
            await ctx.send("Tu n'as pas le droit de participer à ce vote.")
            return
        if not vaction.is_open:
            await ctx.send("Pas de vote pour le nouveau maire en cours !")
            return

        util = vaction.derniere_utilisation

        # Choix de la cible
        candids = CandidHaro.query.filter_by(
            type=CandidHaroType.candidature).all()
        candidats = [candid.joueur.nom for candid in candids]
        pseudo_bc = _BaseCiblageForVote(
            "Pour qui veux-tu voter ? (vote actuel : "
            f"{tools.bold(util.cible.nom if util.cible else 'aucun')})\n"
            f"(candidats : {', '.join(candidats) or 'aucun :pensive:'})"
        )
        cible = await get_cible(ctx, vaction, pseudo_bc, cible)

        # Test si la cible s'est présentée
        cible_ds_candid = CandidHaro.query.filter_by(
            joueur=cible, type=CandidHaroType.candidature).all()
        if not cible_ds_candid:
            mess = await ctx.send(
                f"{cible.nom} ne s'est pas (encore) présenté(e) ! "
                "Si c'est toujours le cas à la fin de l'élection, ton vote "
                "sera compté comme blanc... \n Veux-tu continuer ?"
            )
            if not await tools.yes_no(mess):
                await ctx.send("Compris, mission aborted.")
                return

        if not vaction.is_open:
            # On revérifie, si ça a fermé entre temps !!
            await ctx.send("Le vote pour le nouveau maire a fermé "
                           "entre temps, pas de chance !")
            return

        # Modification en base
        if util.ciblages:       # ancien ciblage
            Ciblage.delete(*util.ciblages)
        Ciblage(utilisation=util, joueur=cible).add()
        util.ts_decision = datetime.datetime.now()
        util.etat = UtilEtat.remplie
        util.update()

        async with ctx.typing():
            # Écriture dans sheet Données brutes
            export_vote(Vote.maire, util)

        await ctx.send(
            f"Vote pour {tools.bold(cible.nom)} bien pris en compte.\n"
            + tools.ital("Tu peux modifier ton vote autant "
                         "que nécessaire avant sa fermeture.")
        )


    @commands.command()
    @tools.vivants_only
    @tools.private
    async def voteloups(self, ctx, *, cible=None):
        """Vote pour la victime de l'attaque des loups

        Args:
            cible: nom du joueur que tu souhaites éliminer.

        Cette commande n'est utilisable que lorsqu'une vote pour la
        victime du soir est en cours, pour les joueurs concernés.

        Le bot t'enverra un message à l'ouverture de chaque vote.

        La commande peut être utilisée autant que voulu pour changer
        de cible tant que le vote est en cours.
        """
        joueur = Joueur.from_member(ctx.author)
        try:
            vaction = joueur.action_vote(Vote.loups)
        except RuntimeError:
            await ctx.send("Minute papillon, le jeu n'est pas encore lancé !")
            return

        # Vérification vote en cours
        if not joueur.votant_loups:
            await ctx.send("Tu n'as pas le droit de participer à ce vote.")
            return
        if not vaction.is_open:
            await ctx.send("Pas de vote pour le nouveau maire en cours !")
            return

        util = vaction.derniere_utilisation

        # Choix de la cible
        pseudo_bc = _BaseCiblageForVote(
            "Qui veux-tu manger ? (vote actuel : "
            f"{tools.bold(util.cible.nom if util.cible else 'aucun')})"
        )
        cible = await get_cible(ctx, vaction, pseudo_bc, cible)

        if not vaction.is_open:
            # On revérifie, si ça a fermé entre temps !!
            await ctx.send("Le vote pour la victime des loups a "
                           "fermé entre temps, pas de chance !")
            return

        # Modification en base
        if util.ciblages:       # ancien ciblage
            Ciblage.delete(*util.ciblages)
        Ciblage(utilisation=util, joueur=cible).add()
        util.ts_decision = datetime.datetime.now()
        util.etat = UtilEtat.remplie
        util.update()

        async with ctx.typing():
            # Écriture dans sheet Données brutes
            export_vote(Vote.loups, util)

        await ctx.send(
            f"Vote contre {tools.bold(cible.nom)} bien pris en compte."
        )


    @commands.command()
    @tools.vivants_only
    @tools.private
    async def action(self, ctx, *, decision=None):
        """Utilise l'action de ton rôle / une des actions associées

        Args:
            decision: ce que tu souhaites faire.

                Dans le cas où tu as plusieurs actions disponibles,
                ce paramètre n'est pas pris en compte pour éviter
                toute ambiguïté.

        Cette commande n'est utilisable que si tu as au moins une action
        ouverte. Action = pouvoir associé à ton rôle, mais aussi
        pouvoirs ponctuels (Lame Vorpale, Chat d'argent...)
        Le bot t'enverra un message à l'ouverture de chaque action.

        La commande peut être utilisée autant que voulu pour changer
        d'action tant que la fenêtre d'action est en cours, SAUF pour
        certaines actions (dites "instantanées") ayant une conséquence
        immédiate (Barbier, Licorne...). Le bot mettra dans ce cas un
        message d'avertissement.
        """
        joueur = Joueur.from_member(ctx.author)

        # Vérification rôle actif
        if not joueur.role_actif:
            await ctx.send(
                "Tu ne peux pas utiliser tes pouvoirs pour le moment !"
            )
            return

        # Détermine la/les actions en cours pour le joueur
        actions = [ac for ac in joueur.actions_actives if ac.is_open]
        if not actions:
            await ctx.send("Aucune action en cours pour toi.")
            return

        elif (N := len(actions)) > 1:
            decision = None
            # Évite de lancer une décision en blind
            # si le joueur a plusieurs actions
            txt = "Tu as plusieurs actions actuellement ouvertes :\n"
            for i in range(N):
                txt += (f" {tools.emoji_chiffre(i+1)} - "
                        f"{tools.code(actions[i].base.slug)}\n")
            message = await ctx.send(txt + "\nPour laquelle veux-tu agir ?")
            i = await tools.choice(message, N)
            action = actions[i-1]

        else:
            action = actions[0]

        util = action.derniere_utilisation

        # Dernière décision et choix annulation/modification
        delete = False
        if util.ciblages and not decision:
            pencil, trash = "\N{PENCIL}", "\N{WASTEBASKET}"
            message = await ctx.send(
                f"Action actuelle : {tools.bold(action.decision)}\n\n"
                f"Souhaites-tu la modifier ({pencil}) ou l'annuler ({trash}) ?"
            )
            delete = await tools.wait_for_react_clic(
                message, emojis={pencil: False, trash: True}
            )

        # Choix de la décision
        cibles = {}
        if not delete:
            for bc in action.base.base_ciblages:        # Triés par priorité
                cibles[bc] = await get_cible(ctx, action, bc, decision)
                decision = None     # si plus de 1 ciblage, vaut pour le 1er

        if not action.is_open:
            # On revérifie, si ça a fermé entre temps !!
            await ctx.send("L'action a fermé entre temps, pas de chance !")
            return

        # Avertissement si action a conséquence instantanée (barbier...)
        if action.base.instant:
            message = await ctx.send(
                "Attention : cette action a une conséquence instantanée ! "
                "Si tu valides, tu ne pourras pas revenir en arrière.\n"
                "Ça part ?"
            )
            if not await tools.yes_no(message):
                await ctx.send("Mission aborted.")
                return

        # Modification en base
        if util.ciblages:     # ancien ciblages
            Ciblage.delete(*util.ciblages)
        for bc, cible in cibles.items():
            cib = Ciblage(utilisation=util, base=bc)
            cib.valeur = cible      # affecte le bon attribut selon le bc.type
        util.ts_decision = datetime.datetime.now()
        util.etat = UtilEtat.remplie
        util.update()

        async with ctx.typing():
            # Écriture dans sheet Données brutes
            export_vote(None, util)

        # Conséquences si action instantanée
        if action.base.instant:
            await gestion_actions.close_action(action)

            await ctx.send(tools.ital(
                f"[Allô {config.Role.mj.mention}, "
                "conséquence instantanée ici !]"
            ))

        else:
            await ctx.send(
                f"Action « {tools.bold(action.decision)} » bien prise "
                f"en compte pour {tools.code(action.base.slug)}.\n"
                + tools.ital("Tu peux modifier ta décision autant que "
                             "nécessaire avant la fin du créneau.")
            )



"""lg-rez / blocs / Ready-check

Classes to raise an exception when accessing attributes not
respecting a given rule

"""


class NotReadyError(RuntimeError):
    """An attribute is tried to be accessed before it is ready.

    Inherits from :exc:`RuntimeError`.
    """
    pass


class _RCDict(dict):
    def __init__(self, _is_ready=None, _class=None, **kwargs):
        super().__init__(**kwargs)
        if _is_ready is None:
            def _is_ready(item):
                return (item is not None)
        self._is_ready = _is_ready
        if _class:
            self._errormsg = f"'{_class.__qualname__}' has no attribute"
        else:
            self._errormsg = "No attribute"

    def _get_raw(self, name):
        # Raw item acess
        try:
            val = super().__getitem__(name)
        except KeyError:
            raise AttributeError(f"{self._errormsg} '{name}'") from None
        return val

    def __getitem__(self, name):
        # Proxy item access
        val = self._get_raw(name)
        ready = self._is_ready(val)
        if not ready:
            raise NotReadyError(f"'{name}' is not ready yet!")
        return val


class _RCMeta(type):
    def __new__(metacls, name, bases, dict, check=None, check_type=None):
        # register directly private/magic names only
        _prv_dict = {name: dict[name] for name in dict if name.startswith('_')}
        return super().__new__(metacls, name, bases, _prv_dict)

    def __init__(cls, name, bases, dict, check=None, check_type=None):
        _prv_dict = {name: dict[name] for name in dict if name.startswith('_')}
        _pub_dict = {name: dict[name] for name in dict
                     if not name.startswith('_')}
        super().__init__(name, bases, _prv_dict)
        _is_ready = check
        if check_type:
            if check:
                def _is_ready(item):
                    return isinstance(item, check_type) and check(item)
            else:
                def _is_ready(item):
                    return isinstance(item, check_type)
        cls._rc_dict = _RCDict(_is_ready=_is_ready, _class=cls, **_pub_dict)

    def __getattr__(cls, name):
        if name.startswith('_'):
            # private/magic name: dont search in ._rc_dict (infinite recursion)
            raise AttributeError(f"'{cls.__name__}' has no attribute '{name}'")
        return cls._rc_dict[name]

    def __setattr__(cls, name, value):
        if name.startswith('_'):
            super().__setattr__(name, value)
        else:
            cls._rc_dict[name] = value

    def __delattr__(cls, name):
        if name.startswith('_'):
            super().__delattr__(name)
        else:
            del cls._rc_dict[name]

    def __iter__(cls):
        return iter(cls._rc_dict)

    def get_raw(cls, attr):
        return cls._rc_dict._get_raw(attr)


class ReadyCheck(metaclass=_RCMeta):
    """Proxy class to prevent accessing not initialised objects.

    When accessing a class attribute, this class:
        - returns its value (classic behavior) if it is evaluated
          *ready* (see below);
        - raises a :exc:`NotReadyError` exception otherwise.

    Subclass this class to implment readiness check on class attributes
    and define "readiness" as needed. By default, attributes are
    considered *not ready* only if their value is ``None``::

        class NotNone(ReadyCheck):
            a = None            # NotNone.a will raise a NotReadyError
            b = <any object>    # NotNone.b will be the given object

    Use ``check_type`` class-definition argument to define readiness based
    on attributes types (using :func:`isinstance()`)::

        class MustBeList(ReadyCheck, check_type=list):
            a = "TDB"           # MustBeList.a will raise a NotReadyError
            b = [1, 2, 3]       # MustBeList.b will be the given list

    Use ``check`` class-definition argument to define custom readiness
    check (``value -> bool`` function)::

        class MustBePositive(ReadyCheck, check=lambda val: val > 0):
            a = 0               # MustBePositive.a will raise a NotReadyError
            b = 37              # MustBePositive.b will be 37

    If both arguments are provided, attribute type will be checked first
    and custom check will be called only for suitable attributes.

    Attributes can be added, modified and deleted the normal way.
    Readiness is evaluated at access time, so changing an attribute's
    value will change its readiness with no aditionnal work required.

    Note:
        Attributes whose name start with ``'_'`` (private and magic attributes)
        are not affected and will be returned even if not ready.

    These classes can also implement iterating protocol to provide access
    to protected attributes **names** (order not reliable)::

        for name in NotNone:
            print(name)         # Will print 'a' then 'b'

    Warning:
        Class derivating from this class are not meant to be instantiated.
        Due to the checking proxy on class attributes, instances will not
        see attributes defined at class level, and attributes defined in
        ``__init__`` or after construction will **not** be ready-checked.

        This class defines no attributes or methods, but relies on a custom
        metaclass: you will not be able to create mixin classes from this
        one and a custom-metaclass one.

    .. py:classmethod:: get_raw(name)

        Access to an attribute bypassing ready check

        :param str name: name of the attribute. This must be a
            **public** attribute (no leading underscore).
        :returns: The attribute value, whatever it is.
        :raises AttributeError: if the attribute doesnt exist
    """
    pass        # Everything is in the metaclass!



import datetime
import unittest
from unittest import mock

import freezegun

from lgrez import config
from lgrez.bdd import model_actions
from lgrez.blocs import webhook
from test import mock_discord, mock_env



class TestAction(unittest.IsolatedAsyncioTestCase):
    """Unit tests for lgrez.bdd.model_actions.Action methods."""

    def test___repr__(self):
        """Unit tests for Action.__repr__ method."""
        # def __repr__(self)
        __repr__ = model_actions.Action.__repr__
        slf = mock.Mock(model_actions.Action, id=11, base="bazz", joueur="j0r")
        rpr = __repr__(slf)
        self.assertEqual(rpr, "<Action #11 (bazz/j0r)>")


class TestTache(unittest.IsolatedAsyncioTestCase):
    """Unit tests for lgrez.bdd.model_actions.Tache methods."""

    def setUp(self):
        mock_discord.mock_config()

    def tearDown(self):
        mock_discord.unmock_config()

    def test___repr__(self):
        """Unit tests for Tache.__repr__ method."""
        # def __repr__(self)
        __repr__ = model_actions.Tache.__repr__
        slf = mock.Mock(model_actions.Tache, id=11, commande="blebelebelpof")
        rpr = __repr__(slf)
        self.assertEqual(rpr, "<Tache #11 (blebelebelpof)>")

    def test_handler(self):
        """Unit tests for Tache.handler property."""
        # @property def handler(self)
        handler = model_actions.Tache.handler
        config.bot.tasks = {11: mock.Mock(), 13: mock.Mock()}
        # Getter - Enregistrée
        slf = mock.Mock(model_actions.Tache, id=11)
        hdlr = handler.fget(slf)
        self.assertEqual(hdlr, config.bot.tasks[11])
        # Getter - Non enregistrée
        slf = mock.Mock(model_actions.Tache, id=12)
        with self.assertRaises(RuntimeError):
            handler.fget(slf)
        # Setter
        value = mock.Mock()
        handler.fset(slf, value)
        self.assertEqual(config.bot.tasks.get(12), value)
        with self.assertRaises(RuntimeError):
            handler.fset(mock.Mock(model_actions.Tache, id=None), value)
        # Deleter - Enregistrée
        handler.fdel(slf)
        self.assertNotIn(12, config.bot.tasks)
        # Deleter - Non enregistrée
        handler.fdel(slf)
        self.assertNotIn(12, config.bot.tasks)


    @mock_env.patch_env(LGREZ_WEBHOOK_URL="uiui://oh")
    @mock.patch("lgrez.blocs.webhook.send")
    def test_execute(self, send_patch):
        """Unit tests for Tache.execute method."""
        # def execute(self)
        execute = model_actions.Tache.execute

        # too soon
        webhook._last_time = 132
        slf = mock.Mock(model_actions.Tache)
        with freezegun.freeze_time(datetime.datetime.utcfromtimestamp(133.5)):
            execute(slf)
        config.loop.call_later.assert_called_once_with(2, slf.execute)
        config.loop.reset_mock()

        # perfect (+2s), envoi OK
        webhook._last_time = 132
        slf = mock.Mock(model_actions.Tache)
        send_patch.return_value = True
        with freezegun.freeze_time(datetime.datetime.utcfromtimestamp(134)):
            execute(slf)
        self.assertEqual(webhook._last_time, 134)
        send_patch.assert_called_once_with(slf.commande, url="uiui://oh")
        send_patch.reset_mock()
        slf.delete.assert_called_once()
        config.loop.call_later.assert_not_called()

        # large, envoi échec
        webhook._last_time = 132
        slf = mock.Mock(model_actions.Tache)
        send_patch.return_value = False
        with freezegun.freeze_time(datetime.datetime.utcfromtimestamp(334)):
            execute(slf)
        self.assertEqual(webhook._last_time, 334)
        send_patch.assert_called_once_with(slf.commande, url="uiui://oh")
        slf.delete.assert_not_called()
        config.loop.call_later.assert_called_once_with(2, slf.execute)


    def test_register(self):
        """Unit tests for Tache.register method."""
        # def register(self)
        register = model_actions.Tache.register

        tache_timestamp = datetime.datetime(2021, 4, 25, 12, 2)
        now_timestamp = datetime.datetime(2021, 3, 22, 2, 20, 4)
        delay = (tache_timestamp - now_timestamp).total_seconds()

        slf = mock.Mock(model_actions.Tache, timestamp=tache_timestamp)
        with freezegun.freeze_time(now_timestamp):
            register(slf)
        config.loop.call_later.assert_called_once_with(delay, slf.execute)
        self.assertEqual(slf.handler, config.loop.call_later.return_value)


    def test_cancel(self):
        """Unit tests for Tache.cancel method."""
        # def cancel(self)
        cancel = model_actions.Tache.cancel
        slf = mock.Mock(model_actions.Tache)
        class EffectSaver:
            called = False
            @classmethod
            def save(cls):
                cls.called = True
        slf.handler = mock.Mock()
        slf.handler.cancel.side_effect = EffectSaver.save
        self.assertTrue(hasattr(slf, "handler"))
        cancel(slf)
        self.assertFalse(hasattr(slf, "handler"))
        self.assertTrue(EffectSaver.called)

        # no handler
        slf = mock.Mock(model_actions.Tache)
        slf.handler.cancel.side_effect = RuntimeError
        cancel(slf)
        self.assertTrue(hasattr(slf, "handler"))


    @mock.patch("lgrez.bdd.base.TableBase.add")
    def test_add(self, sa_patch):
        """Unit tests for Tache.add method."""
        # def add(self, *other)
        add = model_actions.Tache.add
        # one
        slf = mock.Mock(model_actions.Tache)
        add(slf)
        slf.register.assert_called_once()
        sa_patch.assert_called_once()
        sa_patch.reset_mock()
        # several
        slf = mock.Mock(model_actions.Tache)
        other = [mock.Mock(model_actions.Tache) for i in range(5)]
        add(slf, *other)
        slf.register.assert_called_once()
        for oth in other:
            oth.register.assert_called_once()
        sa_patch.assert_called_once_with(*other)


    @mock.patch("lgrez.bdd.base.TableBase.delete")
    def test_delete(self, sd_patch):
        """Unit tests for Tache.delete method."""
        # def delete(self, *other)
        delete = model_actions.Tache.delete
        # one
        slf = mock.Mock(model_actions.Tache)
        delete(slf)
        slf.cancel.assert_called_once()
        sd_patch.assert_called_once()
        sd_patch.reset_mock()
        # several
        slf = mock.Mock(model_actions.Tache)
        other = [mock.Mock(model_actions.Tache) for i in range(5)]
        delete(slf, *other)
        slf.cancel.assert_called_once()
        for oth in other:
            oth.cancel.assert_called_once()
        sd_patch.assert_called_once_with(*other)



import unittest
from unittest import mock

import discord

from lgrez import config, bdd
from lgrez.features import communication
from test import mock_discord, mock_bdd


class TestCommunication(unittest.IsolatedAsyncioTestCase):
    """Unit tests for lgrez.features.communication commands."""

    def setUp(self):
        mock_discord.mock_config()
        self.cog = communication.Communication(config.bot)

    def tearDown(self):
        mock_discord.unmock_config()


    @unittest.SkipTest
    @mock_bdd.patch_db      # Empty database for this method
    async def test_embed(self):
        """Unit tests for !embed command."""
        # async def embed(self, ctx, key=None, *, val=None)
        embed = self.cog.embed


    @mock_bdd.patch_db      # Empty database for this method
    async def test_send(self):
        """Unit tests for !send command."""
        # async def send(self, ctx, cible, *, message)
        send = self.cog.send
        mock_bdd.add_campsroles(10, 10)
        joueurs = [
            bdd.Joueur(discord_id=1, chan_id_=11, nom="Joueur1",
                       statut=bdd.Statut.vivant, _role_slug="role1",
                       votant_village=True, votant_loups=False,
                       role_actif=False),
            bdd.Joueur(discord_id=2, chan_id_=21, nom="Joueur2",
                       chambre="ch2", statut=bdd.Statut.vivant,
                       votant_village=True, votant_loups=False,
                       role_actif=False),
            bdd.Joueur(discord_id=3, chan_id_=31, nom="Joueur3",
                       statut=bdd.Statut.mort, _role_slug="role3",
                       _camp_slug="camp3", votant_village=True,
                       votant_loups=False, role_actif=True),
            bdd.Joueur(discord_id=4, chan_id_=41, nom="Joueur4",
                       chambre="ch 4", statut=bdd.Statut.mort,
                       votant_village=False, votant_loups=False,
                       role_actif=True),
            bdd.Joueur(discord_id=5, chan_id_=51, nom="Joueur 5",
                       statut=bdd.Statut.MV, _camp_slug="camp3",
                       votant_village=False, votant_loups=False,
                       role_actif=False),
            bdd.Joueur(discord_id=6, chan_id_=61, nom="Joueur6",
                       statut=bdd.Statut.immortel, _role_slug="role3",
                       votant_village=False, votant_loups=True,
                       role_actif=False),
        ]
        bdd.Joueur.add(*joueurs)

        # cible = "all"
        ctx = mock_discord.get_ctx(send, "all", message="ouizz")
        with mock_discord.mock_members_and_chans(*joueurs):
            chans = [joueur.private_chan for joueur in joueurs]
            await ctx.invoke()
        ctx.assert_sent("6 trouvé", "Fini")
        for chan in chans:
            chan.send.assert_called_once_with("ouizz")

        # cible = "vivants"
        ctx = mock_discord.get_ctx(send, "vivants", message="nonkk")
        with mock_discord.mock_members_and_chans(*joueurs):
            chans = [joueur.private_chan for joueur in joueurs]
            await ctx.invoke()
        ctx.assert_sent("3 trouvé", "Fini")
        for i, chan in enumerate(chans):
            if i in [0, 1, 4]:
                chan.send.assert_called_once_with("nonkk")
            else:
                chan.send.assert_not_called()

        # cible = "morts"
        ctx = mock_discord.get_ctx(send, "morts", message="bzzk.!")
        with mock_discord.mock_members_and_chans(*joueurs):
            chans = [joueur.private_chan for joueur in joueurs]
            await ctx.invoke()
        ctx.assert_sent("2 trouvé", "Fini")
        for i, chan in enumerate(chans):
            if i in [2, 3]:
                chan.send.assert_called_once_with("bzzk.!")
            else:
                chan.send.assert_not_called()

        # cible = bat crit
        ctx = mock_discord.get_ctx(send, "biz=oui", message="hmm")
        with self.assertRaises(discord.ext.commands.UserInputError) as cm:
            await ctx.invoke()
        self.assertIn("critère 'biz' incorrect", cm.exception.args[0])


        # cible = all crits
        crits = {
            "discord_id=3": [2],
            "chan_id_=21": [1],
            "nom=Joueur4": [3],
            "nom=Joueur 5": [4],
            "chambre=ch2": [1],
            "chambre=ch 4": [3],
            "statut=vivant": [0, 1],
            "statut=mort": [2, 3],
            "statut=MV": [4],
            "statut=immortel": [5],
            "role=role1": [0],
            "role=role3": [2, 5],
            "camp=camp3": [2, 4],
            "votant_village=True": [0, 1, 2],
            "votant_village=1": [0, 1, 2],
            "votant_village=faux": [3, 4, 5],
            "votant_loups=vrai": [5],
            "votant_loups=0": [0, 1, 2, 3, 4],
            "role_actif=True": [2, 3],
            "role_actif=False": [0, 1, 4, 5],
        }
        for crit, ijs in crits.items():
            ctx = mock_discord.get_ctx(send, crit, message="bakka")
            with mock_discord.mock_members_and_chans(*joueurs):
                chans = [joueur.private_chan for joueur in joueurs]
                await ctx.invoke()
            ctx.assert_sent(f"{len(ijs)} trouvé", "Fini")
            for i, chan in enumerate(chans):
                if i in ijs:
                    chan.send.assert_called_once_with("bakka")
                else:
                    chan.send.assert_not_called()

        # cible = non-existing joueur, correct
        ctx = mock_discord.get_ctx(send, "gouzigouzi", message="oui")
        with mock_discord.interact(
            ("wait_for_message_here", mock_discord.message(ctx, "krr")),
            ("wait_for_message_here", mock_discord.message(ctx, "Joueur2")),
        ):
            with mock_discord.mock_members_and_chans(*joueurs):
                chans = [joueur.private_chan for joueur in joueurs]
                await ctx.invoke()
        ctx.assert_sent("Aucune entrée trouvée", "Aucune entrée trouvée",
                        "1 trouvé", "Fini")
        for i, chan in enumerate(chans):
            if i == 1:
                chan.send.assert_called_once_with("oui")
            else:
                chan.send.assert_not_called()

        # cible = existing joueur, correct
        ctx = mock_discord.get_ctx(send, "Joueur 5", message="oui")
        with mock_discord.mock_members_and_chans(*joueurs):
            chans = [joueur.private_chan for joueur in joueurs]
            await ctx.invoke()
        ctx.assert_sent("1 trouvé", "Fini")
        for i, chan in enumerate(chans):
            if i == 4:
                chan.send.assert_called_once_with("oui")
            else:
                chan.send.assert_not_called()

        # cible = all, test eval
        ctx = mock_discord.get_ctx(send, "all", message=
            "Salut {member.mention}, tu t'appelles {joueur.role}, rôle "
            "{joueur.role.nom}, camp {joueur.camp.nom}, chan {chan.name}"
        )
        with mock_discord.mock_members_and_chans(*joueurs):
            members = [joueur.member for joueur in joueurs]
            chans = [joueur.private_chan for joueur in joueurs]
            await ctx.invoke()
        ctx.assert_sent("6 trouvé", "Fini")
        for joueur, chan, member in zip(joueurs, chans, members):
            chan.send.assert_called_once_with(
                f"Salut {member.mention}, tu t'appelles {joueur.role}, rôle "
                f"{joueur.role.nom}, camp {joueur.camp.nom}, chan {chan.name}"
            )


    @mock_bdd.patch_db      # Empty database for this method
    async def test_post(self):
        """Unit tests for !post command."""
        # async def post(self, ctx, chan, *, message)
        post = self.cog.post
        joueurs = [
            bdd.Joueur(discord_id=1, chan_id_=11, nom="Joueur1"),
            bdd.Joueur(discord_id=2, chan_id_=21, nom="Joueur2"),
            bdd.Joueur(discord_id=3, chan_id_=31, nom="Joueur3"),
            bdd.Joueur(discord_id=4, chan_id_=41, nom="Joueur4"),
            bdd.Joueur(discord_id=5, chan_id_=51, nom="Joueur5"),
            bdd.Joueur(discord_id=6, chan_id_=61, nom="Joueur6"),
        ]
        bdd.Joueur.add(*joueurs)

        chans = [mock_discord.chan("logs", id=153447897153447897),
                mock_discord.chan("rôles", id=564894149416486431)]

        # cible = "logs"
        ctx = mock_discord.get_ctx(post, "logs", message="ouizz")
        _chans = config.guild.channels
        config.guild.channels = chans
        await ctx.invoke()
        config.guild.channels = _chans
        ctx.assert_sent("Fait")
        chans[0].send.assert_called_once_with("ouizz")
        chans[0].send.reset_mock()
        chans[1].send.assert_not_called()

        # cible = mention de #logs
        ctx = mock_discord.get_ctx(post, "<#153447897153447897>", message="oz")
        _chans = config.guild.channels
        config.guild.channels = chans
        await ctx.invoke()
        config.guild.channels = _chans
        ctx.assert_sent("Fait")
        chans[0].send.assert_called_once_with("oz")
        chans[0].send.reset_mock()
        chans[1].send.assert_not_called()


    @unittest.SkipTest
    @mock_bdd.patch_db      # Empty database for this method
    async def test_plot(self):
        """Unit tests for !send command."""
        # async def plot(self, ctx, type)
        plot = self.cog.plot


    @mock_bdd.patch_db      # Empty database for this method
    @mock.patch("lgrez.config.Channel.annonces.send")
    async def test_annoncemort(self, as_patch):
        """Unit tests for !annoncemort command."""
        # async def annoncemort(self, ctx, *, victime=None)
        annoncemort = self.cog.annoncemort
        mock_bdd.add_campsroles(10, 10)
        joueurs = [
            bdd.Joueur(discord_id=1, chan_id_=11, nom="Joueur1",
                       statut=bdd.Statut.vivant, _role_slug="role1"),
            bdd.Joueur(discord_id=2, chan_id_=21, nom="Joueur2",
                       chambre="ch2", statut=bdd.Statut.vivant),
            bdd.Joueur(discord_id=3, chan_id_=31, nom="Joueur3",
                       statut=bdd.Statut.mort, _role_slug="role3",
                       _camp_slug="camp3"),
            bdd.Joueur(discord_id=4, chan_id_=41, nom="Joueur4",
                       chambre="ch 4", statut=bdd.Statut.mort),
            bdd.Joueur(discord_id=5, chan_id_=51, nom="Joueur5",
                       statut=bdd.Statut.MV, _role_slug="role5"),
            bdd.Joueur(discord_id=6, chan_id_=61, nom="Joueur6",
                       statut=bdd.Statut.immortel, _role_slug="role3"),
        ]
        bdd.Joueur.add(*joueurs)

        # victime = None, correct "Joueur3", role ok, abort
        ctx = mock_discord.get_ctx(annoncemort)
        with mock_discord.interact(
            ("wait_for_message_here", mock_discord.message(ctx, "bzz")),
            ("wait_for_message_here", mock_discord.message(ctx, "Joueur3")),
            ("yes_no", True),                   # rôle à afficher
            ("wait_for_message_here", mock_discord.message(ctx, "oui")),
            ("yes_no", False),                  # abort
        ):
            await ctx.invoke(cog=self.cog)
        ctx.assert_sent("Qui", "Aucune entrée trouvée", "Rôle à afficher",
                        "Contexte", "Ça part", "Mission aborted")
        as_patch.assert_not_called()

        # victime = "Joueur3", role ok, proceed
        ctx = mock_discord.get_ctx(annoncemort, victime="Joueur3")
        with mock_discord.interact(
            ("yes_no", True),                   # rôle à afficher
            ("wait_for_message_here", mock_discord.message(ctx, "oui")),
            ("yes_no", True),                   # proceed
        ):
            await ctx.invoke(cog=self.cog)
        ctx.assert_sent("Rôle à afficher", "Contexte", "Ça part",
                        "c'est parti")
        as_patch.assert_called_once()
        self.assertIn("quelque chose", as_patch.call_args.args[0])
        embed = as_patch.call_args.kwargs["embed"]
        self.assertIn("Joueur3", embed.title)
        self.assertIn("Role3", embed.title)
        self.assertIn("oui", embed.description)
        self.assertEqual(discord.Embed.Empty, embed.thumbnail.url)
        # standard role: no emoji
        as_patch.reset_mock()

        # victime = "Joueur3", role ok, proceed + emoji
        ctx = mock_discord.get_ctx(annoncemort, victime="Joueur3")
        emoji3 = mock.NonCallableMock(discord.Emoji)
        emoji3.configure_mock(name="emoji3", url="bzooop")
        _emojis = config.guild.emojis
        config.guild.emojis = [emoji3]
        with mock_discord.interact(
            ("yes_no", True),                   # rôle à afficher
            ("wait_for_message_here", mock_discord.message(ctx, "oui")),
            ("yes_no", True),                   # proceed
        ):
            await ctx.invoke(cog=self.cog)
        config.guild.emojis = _emojis
        ctx.assert_sent("Rôle à afficher", "Contexte", "Ça part",
                        "c'est parti")
        as_patch.assert_called_once()
        self.assertIn("quelque chose", as_patch.call_args.args[0])
        embed = as_patch.call_args.kwargs["embed"]
        self.assertIn("Joueur3", embed.title)
        self.assertIn("Role3", embed.title)
        self.assertIn("oui", embed.description)
        self.assertEqual("bzooop", embed.thumbnail.url)
        as_patch.reset_mock()

        # victime = "Joueur5" (MV), role ok no MV, proceed
        ctx = mock_discord.get_ctx(annoncemort, victime="Joueur5")
        with mock_discord.interact(
            ("yes_no", True),                   # rôle à afficher
            ("yes_no", False),                  # afficher la MVance
            ("wait_for_message_here", mock_discord.message(ctx, "oui")),
            ("yes_no", True),                   # proceed
        ):
            await ctx.invoke(cog=self.cog)
        ctx.assert_sent("Rôle à afficher", "Annoncer la mort-vivance",
                         "Contexte", "Ça part", "c'est parti")
        as_patch.assert_called_once()
        self.assertIn("quelque chose", as_patch.call_args.args[0])
        embed = as_patch.call_args.kwargs["embed"]
        self.assertIn("Joueur5", embed.title)
        self.assertIn("Role5", embed.title)
        self.assertNotIn("Mort-Vivant", embed.title)
        self.assertIn("oui", embed.description)
        as_patch.reset_mock()

        # victime = "Joueur5" (MV), role ok no MV, proceed
        ctx = mock_discord.get_ctx(annoncemort, victime="Joueur5")
        with mock_discord.interact(
            ("yes_no", True),                   # rôle à afficher
            ("yes_no", True),                   # afficher la MVance
            ("wait_for_message_here", mock_discord.message(ctx, "oui")),
            ("yes_no", True),                   # proceed
        ):
            await ctx.invoke(cog=self.cog)
        ctx.assert_sent("Rôle à afficher", "Annoncer la mort-vivance",
                         "Contexte", "Ça part", "c'est parti")
        as_patch.assert_called_once()
        self.assertIn("quelque chose", as_patch.call_args.args[0])
        embed = as_patch.call_args.kwargs["embed"]
        self.assertIn("Joueur5", embed.title)
        self.assertIn("Role5", embed.title)
        self.assertIn("Mort-Vivant", embed.title)
        self.assertIn("oui", embed.description)
        as_patch.reset_mock()

        role_village = bdd.Role(slug="")

        # victime = "Joueur3", role not ok usoab, proceed
        ctx = mock_discord.get_ctx(annoncemort, victime="Joueur3")
        with mock_discord.interact(
            ("yes_no", False),                  # rôle à afficher
            ("wait_for_message_here", mock_discord.message(ctx, "RolZZZ")),
            ("wait_for_react_clic", mock.NonCallableMock(discord.Emoji,
                                                         url="bzooop")),
            ("wait_for_message_here", mock_discord.message(ctx, "oui")),
            ("yes_no", True),                   # proceed
        ):
            await ctx.invoke(cog=self.cog)
        ctx.assert_sent("Rôle à afficher", "Rôle à afficher", "Camp",
                        "Contexte", "Ça part", "c'est parti")
        as_patch.assert_called_once()
        self.assertIn("quelque chose", as_patch.call_args.args[0])
        embed = as_patch.call_args.kwargs["embed"]
        self.assertIn("Joueur3", embed.title)
        self.assertIn("RolZZZ", embed.title)
        self.assertIn("oui", embed.description)
        self.assertEqual("bzooop", embed.thumbnail.url)
        as_patch.reset_mock()



import datetime
import functools
import unittest
from unittest import mock

import discord
from discord.ext import commands
import freezegun
import sqlalchemy

from lgrez import bot, config, bdd, blocs, __version__
from lgrez.blocs import ready_check
from test import mock_discord, mock_bdd, mock_env



class TestBotOnReady(unittest.IsolatedAsyncioTestCase):
    """Unit tests for lgrez.bot on_ready function."""

    def setUp(self):
        # partial set up (LGBot.run effects only)
        config.bot = mock.MagicMock(bot.LGBot(), command_prefix="!")
        config.session = mock.Mock(sqlalchemy.orm.session.Session)
        config.engine = mock.Mock(sqlalchemy.engine.Engine)

    def tearDown(self):
        mock_discord.unmock_config()

    @mock_bdd.patch_db()
    @mock.patch("builtins.print")
    @mock.patch("lgrez.blocs.tools.channel")
    @mock.patch("lgrez.blocs.tools.role")
    @mock.patch("lgrez.blocs.tools.emoji")
    async def test__on_ready(self, emo_patch, role_patch, ch_patch, ppatch):
        """Unit tests for bot._on_ready function."""
        # async def _on_ready(bot)
        _on_ready = bot._on_ready

        # guild not found
        config.bot.get_guild.return_value = None
        with self.assertRaises(RuntimeError) as cm:
            await _on_ready(config.bot)
        self.assertEqual(config.loop, config.bot.loop)
        config.bot.i_am_alive.assert_not_called()
        config.bot.get_guild.assert_called_once_with(config.bot.GUILD_ID)
        self.assertIn(str(config.bot.GUILD_ID), cm.exception.args[0])
        ppatch.assert_not_called()
        config.bot.get_guild.reset_mock(return_value=True)
        config.bot.reset_mock()

        # guild found, everything ok
        config.output_liveness = False
        ch_patch.side_effect=functools.partial(mock.Mock, discord.TextChannel)
        role_patch.side_effect=functools.partial(mock.Mock, discord.Role)
        emo_patch.side_effect=functools.partial(mock.Mock, discord.Emoji)
        await _on_ready(config.bot)
        guild = config.bot.get_guild.return_value
        self.assertEqual(config.loop, config.bot.loop)
        config.bot.i_am_alive.assert_not_called()
        config.bot.get_guild.assert_called_once_with(config.bot.GUILD_ID)
        self.assertEqual(config.guild, guild)
        printed = "\n".join(call.args[0] for call in ppatch.call_args_list)
        self.assertIn("Connected", printed)
        self.assertIn(str(guild.name), printed)
        self.assertIn("Initialization", printed)
        self.assertIn("Initialization", printed)
        self.assertIn("Initialization complete", printed)
        self.assertIn("Listening for events", printed)
        mock_discord.assert_sent(config.Channel.logs, "Just rebooted")
        mock_discord.assert_not_sent(config.Channel.logs,
                                     ["tâches planifiées", "ERREURS"])
        ppatch.reset_mock()
        try:
            for name in config.Channel:
                getattr(config.Channel, name)
            for name in config.Role:
                getattr(config.Role, name)
            for name in config.Emoji:
                getattr(config.Emoji, name)
            config.private_chan_category_name
        except ready_check.NotReadyError as e:
            raise AssertionError from e
        config.bot.change_presence.assert_called_once()
        config.bot.reset_mock()

        mock_discord._unprepare_attributes(config.Role)
        mock_discord._unprepare_attributes(config.Channel)
        mock_discord._unprepare_attributes(config.Emoji)

        # guild found, some missing Discord elements (not #logs / @MJ)
        def _channel(name):
            if name in ["rôles", "débats", config.private_chan_category_name]:
                raise ValueError
            else:
                return mock.NonCallableMock(discord.TextChannel)
        def _role(name):
            if name in ["Joueur en vie", "Maire"]:
                raise ValueError
            else:
                return mock.NonCallableMock(discord.Role)
        def _emoji(name):
            if name == "ro":
                raise ValueError
            else:
                return mock.NonCallableMock(discord.Emoji)
        ch_patch.side_effect = _channel
        role_patch.side_effect = _role
        emo_patch.side_effect = _emoji
        await _on_ready(config.bot)
        guild = config.bot.get_guild.return_value
        self.assertEqual(config.loop, config.bot.loop)
        config.bot.i_am_alive.assert_not_called()
        config.bot.get_guild.assert_called_once_with(config.bot.GUILD_ID)
        self.assertEqual(config.guild, guild)
        printed = "\n".join(call.args[0] for call in ppatch.call_args_list)
        self.assertIn("Connected", printed)
        self.assertIn(str(guild.name), printed)
        self.assertIn("Initialization", printed)
        self.assertIn("Initialization complete", printed)
        self.assertIn("Listening for events", printed)
        mock_discord.assert_sent(
            config.Channel.logs,
            ["ERREURS", "6 errors", 'config.Channel.roles = "rôles"',
             'config.Channel.debats = "débats"', str(config.Role.mj.mention),
             'catégorie config.private_chan_category_name = '
             f'"{config.private_chan_category_name}"',
             'config.Role.joueur_en_vie = "Joueur en vie"',
             'config.Role.maire = "Maire"', 'config.Emoji.ro = "ro"'],
            "Just rebooted")
        mock_discord.assert_not_sent(config.Channel.logs, "tâches planifiées",
                                     "tâches planifiées")
        ppatch.reset_mock()
        try:
            for attr in config.Channel:
                if attr not in ["roles", "debats"]:
                    getattr(config.Channel, attr)
            for attr in config.Role:
                if attr not in ["joueur_en_vie", "maire"]:
                    getattr(config.Role, attr)
            for attr in config.Emoji:
                if attr != "ro":
                    getattr(config.Emoji, attr)
            config.private_chan_category_name
        except ready_check.NotReadyError as e:
            raise AssertionError from e
        config.bot.change_presence.assert_called_once()
        config.bot.reset_mock()

        mock_discord._unprepare_attributes(config.Role)
        mock_discord._unprepare_attributes(config.Channel)
        mock_discord._unprepare_attributes(config.Emoji)

        # guild found, missing @MJ
        def _channel(name):
            return mock.NonCallableMock(discord.TextChannel)
        def _role(name):
            if name == "MJ":
                raise ValueError
            else:
                return mock.NonCallableMock(discord.Role)
        def _emoji(name):
            return mock.NonCallableMock(discord.Emoji)
        ch_patch.side_effect = _channel
        role_patch.side_effect = _role
        emo_patch.side_effect = _emoji
        await _on_ready(config.bot)
        guild = config.bot.get_guild.return_value
        self.assertEqual(config.loop, config.bot.loop)
        config.bot.i_am_alive.assert_not_called()
        config.bot.get_guild.assert_called_once_with(config.bot.GUILD_ID)
        self.assertEqual(config.guild, guild)
        printed = "\n".join(call.args[0] for call in ppatch.call_args_list)
        self.assertIn("Connected", printed)
        self.assertIn(str(guild.name), printed)
        self.assertIn("Initialization", printed)
        self.assertIn("Initialization complete", printed)
        self.assertIn("Listening for events", printed)
        mock_discord.assert_sent(
            config.Channel.logs,
            ["ERREURS", "1 error", 'config.Role.mj = "MJ"',
             "@everyone"],
            "Just rebooted")
        mock_discord.assert_not_sent(config.Channel.logs, "tâches planifiées",
                                     "tâches planifiées")
        ppatch.reset_mock()
        try:
            for attr in config.Channel:
                getattr(config.Channel, attr)
            for attr in config.Role:
                if attr != "mj":
                    getattr(config.Role, attr)
            for attr in config.Emoji:
                getattr(config.Emoji, attr)
            config.private_chan_category_name
        except ready_check.NotReadyError as e:
            raise AssertionError from e
        config.bot.change_presence.assert_called_once()
        config.bot.reset_mock()

        mock_discord._unprepare_attributes(config.Role)
        mock_discord._unprepare_attributes(config.Channel)
        mock_discord._unprepare_attributes(config.Emoji)

        # guild found, missing @MJ / #logs
        def _channel(name):
            if name == "logs":
                raise ValueError
            else:
                return mock.NonCallableMock(discord.TextChannel)
        def _role(name):
            return mock.NonCallableMock(discord.Role)
        def _emoji(name):
            return mock.NonCallableMock(discord.Emoji)
        ch_patch.side_effect = _channel
        role_patch.side_effect = _role
        emo_patch.side_effect = _emoji
        defchan = mock.NonCallableMock(discord.TextChannel)
        config.guild.text_channels.__getitem__.return_value = defchan
        await _on_ready(config.bot)
        guild = config.bot.get_guild.return_value
        self.assertEqual(config.loop, config.bot.loop)
        config.bot.i_am_alive.assert_not_called()
        config.bot.get_guild.assert_called_once_with(config.bot.GUILD_ID)
        self.assertEqual(config.guild, guild)
        printed = "\n".join(call.args[0] for call in ppatch.call_args_list)
        self.assertIn("Connected", printed)
        self.assertIn(str(guild.name), printed)
        self.assertIn("Initialization", printed)
        self.assertIn("Initialization complete", printed)
        self.assertIn("Listening for events", printed)
        mock_discord.assert_sent(
            defchan,
            ["ERREURS", "1 error", 'config.Channel.logs = "logs"',
             "Routing logs", str(config.Role.mj.mention)],
            "Just rebooted")
        ppatch.reset_mock()
        try:
            for attr in config.Channel:
                if attr != "logs":
                    getattr(config.Channel, attr)
            for attr in config.Role:
                getattr(config.Role, attr)
            for attr in config.Emoji:
                getattr(config.Emoji, attr)
            config.private_chan_category_name
        except ready_check.NotReadyError as e:
            raise AssertionError from e
        config.bot.change_presence.assert_called_once()
        config.bot.reset_mock()

        mock_discord._unprepare_attributes(config.Role)
        mock_discord._unprepare_attributes(config.Channel)
        mock_discord._unprepare_attributes(config.Emoji)

        # guild found, missing @MJ / #logs
        def _channel(name):
            return mock.NonCallableMock(discord.TextChannel)
        def _role(name):
            return mock.NonCallableMock(discord.Role)
        def _emoji(name):
            return mock.NonCallableMock(discord.Emoji)
        taches = [
            bdd.Tache(timestamp=datetime.datetime.now(), commande="cmdz"),
            bdd.Tache(timestamp=datetime.datetime.now(), commande="cmd2"),
            bdd.Tache(timestamp=datetime.datetime.now(), commande="cmd3"),
        ]
        bdd.Tache.add(*taches)
        ch_patch.side_effect = _channel
        role_patch.side_effect = _role
        emo_patch.side_effect = _emoji
        await _on_ready(config.bot)
        guild = config.bot.get_guild.return_value
        self.assertEqual(config.loop, config.bot.loop)
        config.bot.i_am_alive.assert_not_called()
        config.bot.get_guild.assert_called_once_with(config.bot.GUILD_ID)
        self.assertEqual(config.guild, guild)
        printed = "\n".join(call.args[0] for call in ppatch.call_args_list)
        self.assertIn("Connected", printed)
        self.assertIn(str(guild.name), printed)
        self.assertIn("Initialization", printed)
        self.assertIn("Initialization complete", printed)
        self.assertIn("Listening for events", printed)
        mock_discord.assert_sent(config.Channel.logs, "Just rebooted",
                                 "3 tâches planifiées")
        ppatch.reset_mock()
        try:
            for attr in config.Channel:
                if attr != "logs":
                    getattr(config.Channel, attr)
            for attr in config.Role:
                getattr(config.Role, attr)
            for attr in config.Emoji:
                getattr(config.Emoji, attr)
            config.private_chan_category_name
        except ready_check.NotReadyError as e:
            raise AssertionError from e
        config.bot.change_presence.assert_called_once()
        config.loop.call_later.assert_has_calls(
            [mock.call(mock.ANY, tache.execute) for tache in taches]
        )


class TestBotEventsFunction(unittest.IsolatedAsyncioTestCase):
    """Unit tests for lgrez.bot events functions."""

    def setUp(self):
        mock_discord.mock_config()

    def tearDown(self):
        mock_discord.unmock_config()


    @mock.patch("lgrez.features.inscription.main")
    async def test__on_member_join(self, inscr_patch):
        """Unit tests for bot._on_member_join function."""
        # async def _on_member_join(bot, member)
        _on_member_join = bot._on_member_join

        # bad guild
        member = mock.MagicMock(discord.Member, guild=3)
        await _on_member_join(config.bot, member)
        inscr_patch.assert_not_called()

        # good guild
        member = mock.MagicMock(discord.Member, guild=config.guild)
        await _on_member_join(config.bot, member)
        inscr_patch.assert_called_once_with(member)


    @mock.patch("lgrez.blocs.tools.mention_MJ")
    async def test__on_member_remove(self, mmj_patch):
        """Unit tests for bot._on_member_remove function."""
        # async def _on_member_remove(bot, member)
        _on_member_remove = bot._on_member_remove

        # bad guild
        member = mock.MagicMock(discord.Member, guild=3)
        await _on_member_remove(config.bot, member)
        config.Channel.logs.send.assert_not_called()
        mmj_patch.assert_not_called()

        # good guild
        member = mock.MagicMock(discord.Member, guild=config.guild)
        await _on_member_remove(config.bot, member)
        mock_discord.assert_sent(config.Channel.logs,
                                 [mmj_patch.return_value, "ALERTE",
                                  member.display_name])


    @mock.patch("lgrez.features.IA.process_IA")
    async def test__on_message(self, ia_patch):
        """Unit tests for bot._on_message function."""
        # async def _on_message(bot, message)
        _on_message = bot._on_message
        config.bot.get_context = mock.AsyncMock()
        config.bot.invoke = mock.AsyncMock()
        config.bot.in_command = [12]
        config.bot.in_stfu = [13]

        # author = bot
        message = mock.NonCallableMock(discord.Message, author=config.bot.user)
        await _on_message(config.bot, message)
        self.assertEqual(len(message.mock_calls), 0)
        ia_patch.assert_not_called()

        # pas de serveur (MP)
        message = mock.NonCallableMock(discord.Message, guild=None)
        message.channel = mock_discord.chan("some-private-chan")
        await _on_message(config.bot, message)
        message.channel.send.assert_called_once()
        self.assertEqual(len(message.mock_calls), 1)
        ia_patch.assert_not_called()

        # mauvais serveur
        message = mock.NonCallableMock(discord.Message, guild=3)
        await _on_message(config.bot, message)
        self.assertEqual(len(message.mock_calls), 0)
        ia_patch.assert_not_called()

        # pas de rôle affecté
        message = mock.NonCallableMock(discord.Message, guild=config.guild,
                                       webhook_id=None)
        message.author.top_role = config.Role.everyone
        await _on_message(config.bot, message)
        self.assertEqual(len(message.mock_calls), 0)
        ia_patch.assert_not_called()

        # basic
        message = mock.NonCallableMock(
            discord.Message, guild=config.guild, content="bzzt",
            channel=mock_discord.chan("conv-bot-A", id=11)
        )
        await _on_message(config.bot, message)
        config.bot.get_context.assert_called_with(message)
        config.bot.invoke.assert_called_with(
            config.bot.get_context.return_value)
        ia_patch.assert_called_with(message)
        ia_patch.reset_mock()

        # command
        message = mock.NonCallableMock(discord.Message, guild=config.guild,
                                       content="!bzzt")
        await _on_message(config.bot, message)
        config.bot.get_context.assert_called_with(message)
        config.bot.invoke.assert_called_with(
            config.bot.get_context.return_value)
        ia_patch.assert_not_called()

        # command + blanc
        message = mock.NonCallableMock(discord.Message, guild=config.guild,
                                       content="! bzzt")
        await _on_message(config.bot, message)
        config.bot.get_context.assert_called_with(message)
        config.bot.invoke.assert_called_with(
            config.bot.get_context.return_value)
        self.assertEqual(message.content, "!bzzt")
        ia_patch.assert_not_called()

        # autre préfixe
        config.bot.command_prefix = "<OP>"
        message = mock.NonCallableMock(discord.Message, guild=config.guild,
                                       content="<OP>bzzt")
        await _on_message(config.bot, message)
        config.bot.get_context.assert_called_with(message)
        config.bot.invoke.assert_called_with(
            config.bot.get_context.return_value)
        ia_patch.assert_not_called()
        config.bot.command_prefix = "!"

        # not in private chan
        message = mock.NonCallableMock(
            discord.Message, guild=config.guild, content="bzzt",
            channel=mock_discord.chan("gzzt", id=11)
        )
        await _on_message(config.bot, message)
        config.bot.get_context.assert_called_with(message)
        config.bot.invoke.assert_called_with(
            config.bot.get_context.return_value)
        ia_patch.assert_not_called()

        # in private chan - other prefix
        config.private_chan_prefix = "<OP>"
        message = mock.NonCallableMock(
            discord.Message, guild=config.guild, content="bzzt",
            channel=mock_discord.chan("<OP>gzzt", id=11)
        )
        await _on_message(config.bot, message)
        config.bot.get_context.assert_called_with(message)
        config.bot.invoke.assert_called_with(
            config.bot.get_context.return_value)
        ia_patch.assert_called_with(message)
        ia_patch.reset_mock()
        config.private_chan_prefix = "conv-bot-"

        # in command
        message = mock.NonCallableMock(
            discord.Message, guild=config.guild, content="bzzt",
            channel=mock_discord.chan("conv-bot-gzzt", id=12)
        )
        await _on_message(config.bot, message)
        config.bot.get_context.assert_called_with(message)
        config.bot.invoke.assert_called_with(
            config.bot.get_context.return_value)
        ia_patch.assert_not_called()

        # in stfu
        message = mock.NonCallableMock(
            discord.Message, guild=config.guild, content="bzzt",
            channel=mock_discord.chan("conv-bot-gzzt", id=13)
        )
        await _on_message(config.bot, message)
        config.bot.get_context.assert_called_with(message)
        config.bot.invoke.assert_called_with(
            config.bot.get_context.return_value)
        ia_patch.assert_not_called()

        # good
        message = mock.NonCallableMock(
            discord.Message, guild=config.guild, content="bzzt",
            channel=mock_discord.chan("conv-bot-gzzt", id=11)
        )
        await _on_message(config.bot, message)
        config.bot.get_context.assert_called_with(message)
        config.bot.invoke.assert_called_with(
            config.bot.get_context.return_value)
        ia_patch.assert_called_with(message)


    @mock.patch("lgrez.blocs.tools.create_context")
    async def test__on_raw_reaction_add(self, cc_patch):
        """Unit tests for bot._on_raw_reaction_add function."""
        # async def _on_raw_reaction_add(bot, payload)
        _on_raw_reaction_add = bot._on_raw_reaction_add

        # author = bot
        payload = mock.NonCallableMock(discord.RawReactionActionEvent,
                                       member=config.bot.user)
        await _on_raw_reaction_add(config.bot, payload)
        config.bot.invoke.assert_not_called()
        cc_patch.assert_not_called()

        # mauvais serveur
        payload = mock.NonCallableMock(discord.RawReactionActionEvent,
                                       guild_id=10)
        await _on_raw_reaction_add(config.bot, payload)
        config.bot.invoke.assert_not_called()
        cc_patch.assert_not_called()

        # chan non privé
        payload = mock.NonCallableMock(discord.RawReactionActionEvent,
                                       guild_id=config.guild.id)
        config.guild.get_channel.return_value = mock_discord.chan("bzzt")
        await _on_raw_reaction_add(config.bot, payload)
        config.guild.get_channel.assert_called_with(payload.channel_id)
        config.bot.invoke.assert_not_called()
        cc_patch.assert_not_called()

        # pas joueur en vie
        payload = mock.NonCallableMock(
            discord.RawReactionActionEvent, guild_id=config.guild.id,
            member=mock.Mock(discord.Member, roles=[config.Role.joueur_mort])
        )
        config.guild.get_channel.return_value = mock_discord.chan(
            "conv-bot-A")
        await _on_raw_reaction_add(config.bot, payload)
        config.guild.get_channel.assert_called_with(payload.channel_id)
        config.bot.invoke.assert_not_called()
        cc_patch.assert_not_called()

        # bûcher
        payload = mock.NonCallableMock(
            discord.RawReactionActionEvent, guild_id=config.guild.id,
            emoji=config.Emoji.bucher,
            member=mock.Mock(discord.Member, roles=[config.Role.joueur_en_vie])
        )
        chan = mock_discord.chan("conv-bot-A")
        config.guild.get_channel.return_value = chan
        await _on_raw_reaction_add(config.bot, payload)
        config.guild.get_channel.assert_called_with(payload.channel_id)
        config.bot.invoke.assert_called_once()
        cc_patch.assert_called_once_with(payload.member, "!vote")
        sent = cc_patch.return_value.send.call_args.args[0]
        self.assertIn(str(config.Emoji.bucher), sent)
        self.assertIn("Vote pour le condamné", sent)
        config.bot.invoke.reset_mock()
        cc_patch.reset_mock()

        # maire
        payload = mock.NonCallableMock(
            discord.RawReactionActionEvent, guild_id=config.guild.id,
            emoji=config.Emoji.maire,
            member=mock.Mock(discord.Member, roles=[config.Role.joueur_en_vie])
        )
        chan = mock_discord.chan("conv-bot-A")
        config.guild.get_channel.return_value = chan
        await _on_raw_reaction_add(config.bot, payload)
        config.guild.get_channel.assert_called_with(payload.channel_id)
        config.bot.invoke.assert_called_once()
        cc_patch.assert_called_once_with(payload.member, "!votemaire")
        sent = cc_patch.return_value.send.call_args.args[0]
        self.assertIn(str(config.Emoji.maire), sent)
        self.assertIn("Vote pour le nouveau maire", sent)
        config.bot.invoke.reset_mock()
        cc_patch.reset_mock()

        # loups
        payload = mock.NonCallableMock(
            discord.RawReactionActionEvent, guild_id=config.guild.id,
            emoji=config.Emoji.lune,
            member=mock.Mock(discord.Member, roles=[config.Role.joueur_en_vie])
        )
        chan = mock_discord.chan("conv-bot-A")
        config.guild.get_channel.return_value = chan
        await _on_raw_reaction_add(config.bot, payload)
        config.guild.get_channel.assert_called_with(payload.channel_id)
        config.bot.invoke.assert_called_once()
        cc_patch.assert_called_once_with(payload.member, "!voteloups")
        sent = cc_patch.return_value.send.call_args.args[0]
        self.assertIn(str(config.Emoji.lune), sent)
        self.assertIn("Vote pour la victime", sent)
        config.bot.invoke.reset_mock()
        cc_patch.reset_mock()

        # action
        payload = mock.NonCallableMock(
            discord.RawReactionActionEvent, guild_id=config.guild.id,
            emoji=config.Emoji.action,
            member=mock.Mock(discord.Member, roles=[config.Role.joueur_en_vie])
        )
        chan = mock_discord.chan("conv-bot-A")
        config.guild.get_channel.return_value = chan
        await _on_raw_reaction_add(config.bot, payload)
        config.guild.get_channel.assert_called_with(payload.channel_id)
        config.bot.invoke.assert_called_once()
        cc_patch.assert_called_once_with(payload.member, "!action")
        sent = cc_patch.return_value.send.call_args.args[0]
        self.assertIn(str(config.Emoji.action), sent)
        self.assertIn("Action", sent)
        config.bot.invoke.reset_mock()
        cc_patch.reset_mock()

        # other emoji
        payload = mock.NonCallableMock(
            discord.RawReactionActionEvent, guild_id=config.guild.id,
            emoji=config.Emoji.ha,
            member=mock.Mock(discord.Member, roles=[config.Role.joueur_en_vie])
        )
        chan = mock_discord.chan("conv-bot-A")
        config.guild.get_channel.return_value = chan
        await _on_raw_reaction_add(config.bot, payload)
        config.guild.get_channel.assert_called_with(payload.channel_id)
        config.bot.invoke.assert_not_called()
        cc_patch.assert_not_called()


    async def test__on_command_error(self):
        """Unit tests for bot._on_command_error function."""
        # async def _on_command_error(bot, ctx, exc)
        _on_command_error = bot._on_command_error

        # mauvais serveur
        ctx = mock.NonCallableMock(commands.Context, guild=10)
        await _on_command_error(config.bot, ctx, None)
        ctx.send.assert_not_called()

        # STOP
        ctx = mock_discord.get_ctx()
        exc = commands.CommandInvokeError(blocs.tools.CommandExit())
        await _on_command_error(config.bot, ctx, exc)
        ctx.assert_sent("Mission aborted")

        # STOP custom message
        ctx = mock_discord.get_ctx()
        exc = commands.CommandInvokeError(blocs.tools.CommandExit("cust0m"))
        await _on_command_error(config.bot, ctx, exc)
        ctx.assert_sent("cust0m")

        # BDD error - MJ
        ctx = mock_discord.get_ctx()
        ctx.author.top_role = mock.MagicMock(discord.Role,
                                             __ge__=lambda s, o: True) # >= MJ
        exc = commands.CommandInvokeError(bdd.SQLAlchemyError("bzzt"))
        try: raise exc.original      # creating traceback
        except bdd.SQLAlchemyError: pass
        await _on_command_error(config.bot, ctx, exc)
        mock_discord.assert_sent(config.Channel.logs, "Rollback session")
        config.Channel.logs.send.reset_mock()
        config.session.rollback.assert_called_once()
        config.session.rollback.reset_mock()
        ctx.assert_sent(["Un problème est survenu", "Traceback",
                         "SQLAlchemyError", "bzzt"])

        # BDD error - not MJ
        ctx = mock_discord.get_ctx()
        ctx.author.top_role = mock.MagicMock(discord.Role,
                                             __ge__=lambda s, o: False) # < MJ
        exc = commands.CommandInvokeError(bdd.DriverOperationalError("bzoozt"))
        try: raise exc.original      # creating traceback
        except bdd.DriverOperationalError: pass
        await _on_command_error(config.bot, ctx, exc)
        mock_discord.assert_sent(config.Channel.logs, "Rollback session")
        config.Channel.logs.send.reset_mock()
        config.session.rollback.assert_called_once()
        config.session.rollback.reset_mock()
        ctx.assert_sent(["Un problème est survenu",
                         bdd.DriverOperationalError.__name__, "bzoozt"])
        ctx.assert_not_sent("Traceback")

        # BDD error - session not ready : on vérifie juste que pas d'erreur
        ctx = mock_discord.get_ctx()
        _session = config.session
        del config.session
        ctx.author.top_role = mock.MagicMock(discord.Role,
                                             __ge__=lambda s, o: True) # >= MJ
        exc = commands.CommandInvokeError(bdd.SQLAlchemyError("bzzt"))
        await _on_command_error(config.bot, ctx, exc)
        mock_discord.assert_sent(config.Channel.logs)   # nothing
        config.Channel.logs.send.reset_mock()
        ctx.assert_sent(["Un problème est survenu", "SQLAlchemyError", "bzzt"])
        config.session = _session

        # CommandNotFound
        ctx = mock_discord.get_ctx()
        await _on_command_error(config.bot, ctx, commands.CommandNotFound())
        ctx.assert_sent("je ne connais pas cette commande")

        # DisabledCommand
        ctx = mock_discord.get_ctx()
        await _on_command_error(config.bot, ctx, commands.DisabledCommand())
        ctx.assert_sent("Cette commande est désactivée")

        # ConversionError
        command = mock.Mock()
        ctx = mock_discord.get_ctx(command)
        exc = commands.ConversionError("bkrkr", "kak")
        await _on_command_error(config.bot, ctx, exc)
        ctx.assert_sent(["ce n'est pas comme ça qu'on utilise cette commande",
                         "ConversionError", "bkrkr", "kak"])
        config.bot.get_context.assert_called_with(ctx.message)
        self.assertEqual(ctx.message.content, f"!help {command.name}")
        config.bot.get_context.return_value.reinvoke.assert_called()

        # UserInputError
        command = mock.Mock()
        ctx = mock_discord.get_ctx(command)
        exc = commands.UserInputError("bakaka")
        await _on_command_error(config.bot, ctx, exc)
        ctx.assert_sent(["ce n'est pas comme ça qu'on utilise cette commande",
                         "UserInputError", "bakaka"])
        config.bot.get_context.assert_called_with(ctx.message)
        self.assertEqual(ctx.message.content, f"!help {command.name}")
        config.bot.get_context.return_value.reinvoke.assert_called()

        # UserInputError derivate
        command = mock.Mock()
        ctx = mock_discord.get_ctx(command)
        exc = commands.BadArgument("bakaka")
        await _on_command_error(config.bot, ctx, exc)
        ctx.assert_sent(["ce n'est pas comme ça qu'on utilise cette commande",
                         "BadArgument", "bakaka"])
        config.bot.get_context.assert_called_with(ctx.message)
        self.assertEqual(ctx.message.content, f"!help {command.name}")
        config.bot.get_context.return_value.reinvoke.assert_called()

        # CheckAnyFailure
        ctx = mock_discord.get_ctx()
        exc = commands.CheckAnyFailure(mock.ANY, mock.ANY)
        await _on_command_error(config.bot, ctx, exc)
        ctx.assert_sent("cette commande est réservée aux MJs")

        # MissingAnyRole
        ctx = mock_discord.get_ctx()
        exc = commands.MissingAnyRole([mock.ANY])
        await _on_command_error(config.bot, ctx, exc)
        ctx.assert_sent("Cette commande est réservée aux joueurs")

        # MissingRole
        ctx = mock_discord.get_ctx()
        exc = commands.MissingRole(mock.ANY)
        await _on_command_error(config.bot, ctx, exc)
        ctx.assert_sent("cette commande est réservée aux joueurs en vie")

        # one_command.AlreadyInCommand, not addIA/modifIA
        command = mock.Mock()
        command.configure_mock(name="blabla")
        ctx = mock_discord.get_ctx(command)
        exc = blocs.one_command.AlreadyInCommand()
        await _on_command_error(config.bot, ctx, exc)
        ctx.assert_sent("Impossible d'utiliser une commande pendant")

        # one_command.AlreadyInCommand, addIA
        command = mock.Mock()
        command.configure_mock(name="addIA")
        ctx = mock_discord.get_ctx(command)
        exc = blocs.one_command.AlreadyInCommand()
        await _on_command_error(config.bot, ctx, exc)
        ctx.assert_sent()

        # one_command.AlreadyInCommand, modifIA
        command = mock.Mock()
        command.configure_mock(name="modifIA")
        ctx = mock_discord.get_ctx(command)
        exc = blocs.one_command.AlreadyInCommand()
        await _on_command_error(config.bot, ctx, exc)
        ctx.assert_sent()

        # CheckFailure (autre check erreur)
        ctx = mock_discord.get_ctx()
        exc = commands.CheckFailure("jojo")
        with mock.patch("lgrez.blocs.tools.mention_MJ") as mmj_patch:
            await _on_command_error(config.bot, ctx, exc)
        mmj_patch.assert_called_once_with(ctx)
        ctx.assert_sent(["cette commande ne puisse pas être exécutée",
                         "CheckFailure", "jojo"])

        # other exception
        ctx = mock_discord.get_ctx()
        exc = AttributeError("oh")
        with mock.patch("lgrez.blocs.tools.mention_MJ") as mmj_patch:
            await _on_command_error(config.bot, ctx, exc)
        mmj_patch.assert_called_once_with(ctx)
        ctx.assert_sent(["Une erreur inattendue est survenue",
                         "AttributeError", "oh"])

        # other exception
        ctx = mock_discord.get_ctx()
        class SomeException(Exception):
            pass
        exc = SomeException("p!wet")
        with mock.patch("lgrez.blocs.tools.mention_MJ") as mmj_patch:
            await _on_command_error(config.bot, ctx, exc)
        mmj_patch.assert_called_once_with(ctx)
        ctx.assert_sent(["Une erreur inattendue est survenue",
                         "SomeException", "p!wet"])


    async def test__on_error(self):
        """Unit tests for bot._on_error function."""
        # async def _on_error(bot, event, *args, **kwargs)
        _on_error = bot._on_error

        # SQLAlchemyError
        exc = bdd.SQLAlchemyError("bzzt")
        with self.assertRaises(bdd.SQLAlchemyError):    # re-raised
            try:
                raise exc
            except bdd.SQLAlchemyError:
                # called within a running except block
                await _on_error(config.bot, "event")
        mock_discord.assert_sent(
            config.Channel.logs, "Rollback session",
            [f"{config.Role.mj.mention} ALED : Exception Python",
             "Traceback", "SQLAlchemyError", "bzzt"])
        config.Channel.logs.send.reset_mock()
        config.session.rollback.assert_called_once()
        config.session.rollback.reset_mock()

        # DriverOperationalError
        exc = bdd.DriverOperationalError("koook")
        with self.assertRaises(bdd.DriverOperationalError):    # re-raised
            try:
                raise exc
            except bdd.DriverOperationalError:
                # called within a running except block
                await _on_error(config.bot, "event")
        mock_discord.assert_sent(
            config.Channel.logs, "Rollback session",
            [f"{config.Role.mj.mention} ALED : Exception Python",
             "Traceback", bdd.DriverOperationalError.__name__, "koook"])
        config.Channel.logs.send.reset_mock()
        config.session.rollback.assert_called_once()
        config.session.rollback.reset_mock()

        # BDD error - session not ready : on vérifie juste que pas d'erreur
        exc = bdd.SQLAlchemyError("bzzt")
        _session = config.session
        del config.session
        with self.assertRaises(bdd.SQLAlchemyError):    # re-raised
            try:
                raise exc
            except bdd.SQLAlchemyError:
                # called within a running except block
                await _on_error(config.bot, "event")
        mock_discord.assert_sent(
            config.Channel.logs,
            [f"{config.Role.mj.mention} ALED : Exception Python",
             "Traceback", "SQLAlchemyError", "bzzt"])
        config.Channel.logs.send.reset_mock()
        config.session = _session

        # other exception
        class SomeException(Exception):
            pass
        with self.assertRaises(SomeException):    # re-raised
            try:
                raise SomeException("!a!", "@b@", "^c^")
            except SomeException:
                # called within a running except block
                await _on_error(config.bot, "event")
        mock_discord.assert_sent(
            config.Channel.logs,
            [f"{config.Role.mj.mention} ALED : Exception Python",
             "Traceback", "SomeException", "!a!", "@b@", "^c^"])
        config.Channel.logs.send.reset_mock()



class TestLGBot(unittest.IsolatedAsyncioTestCase):
    """Unit tests for lgrez.LGBot class methods."""

    # def setUp(self):
    #     mock_discord.mock_config()
    #
    # def tearDown(self):
    #     mock_discord.unmock_config()


    @mock.patch("lgrez.features.informations.Informations")
    @mock.patch("lgrez.features.voter_agir.VoterAgir")
    @mock.patch("lgrez.features.actions_publiques.ActionsPubliques")
    @mock.patch("lgrez.features.open_close.OpenClose")
    @mock.patch("lgrez.features.sync.Sync")
    @mock.patch("lgrez.features.taches.GestionTaches")
    @mock.patch("lgrez.features.communication.Communication")
    @mock.patch("lgrez.features.IA.GestionIA")
    @mock.patch("lgrez.features.annexe.Annexe")
    @mock.patch("lgrez.features.special.Special")
    async def test___init__(self, *cog_patches):
        """Unit tests for bot.__init__ function."""
        # def __init__(self, command_prefix="!", case_insensitive=True,
        #              description=None, intents=None, member_cache_flags=None,
        #              **kwargs)
        __init__ = bot.LGBot.__init__

        # all default
        created = mock.MagicMock(bot.LGBot)
        with mock.patch("discord.ext.commands.Bot.__init__") as bot_patch:
            __init__(created)
        bot_patch.assert_called_once_with(
            command_prefix="!", description=bot.default_descr,
            case_insensitive=True, intents=discord.Intents.all(),
            member_cache_flags=discord.MemberCacheFlags.all()
        )

        self.assertIsNone(created.GUILD_ID)
        self.assertEqual(created.in_stfu, [])
        self.assertEqual(created.in_fals, [])
        self.assertEqual(created.tasks, {})
        self.assertEqual(created.in_command, [])
        created.add_check.assert_called_once_with(
            blocs.one_command.not_in_command)
        created.before_invoke.assert_called_once_with(
            blocs.one_command.add_to_in_command)
        created.after_invoke.assert_called_once_with(
            blocs.one_command.remove_from_in_command)
        created.remove_command.assert_called_once_with("help")
        for patch in cog_patches:
            patch.assert_called_once_with(created)
        self.assertEqual(
            {call.args[0] for call in created.add_cog.call_args_list},
            {patch.return_value for patch in cog_patches},
        )

        # all custom
        created = mock.MagicMock(bot.LGBot)
        with mock.patch("discord.ext.commands.Bot.__init__") as bot_patch:
            __init__(created, command_prefix="<gez>", case_insensitive=False,
                     description="@descr@", intents=35, member_cache_flags=12,
                     aaa=15, blblbl="oh")
        bot_patch.assert_called_once_with(
            command_prefix="<gez>", description="@descr@",
            case_insensitive=False, intents=35, member_cache_flags=12,
            aaa=15, blblbl="oh"
        )


    @mock.patch("lgrez.bot._on_ready")
    async def test_on_ready(self, ptch):
        """Unit tests for bot.on_ready function."""
        # async def on_ready(self), on_ready(self)
        bt = mock.Mock()
        await bot.LGBot.on_ready(bt)
        ptch.assert_called_once_with(bt)

    @mock.patch("lgrez.bot._on_member_join")
    async def test_on_member_join(self, ptch):
        """Unit tests for bot.on_member_join function."""
        # async def on_member_join(bt, member)
        bt, member = mock.Mock(), mock.Mock()
        await bot.LGBot.on_member_join(bt, member)
        ptch.assert_called_once_with(bt, member)

    @mock.patch("lgrez.bot._on_member_remove")
    async def test_on_member_remove(self, ptch):
        """Unit tests for bot.on_member_remove function."""
        # async def on_member_remove(bt, member)
        bt, member = mock.Mock(), mock.Mock()
        await bot.LGBot.on_member_remove(bt, member)
        ptch.assert_called_once_with(bt, member)

    @mock.patch("lgrez.bot._on_message")
    async def test_on_message(self, ptch):
        """Unit tests for bot.on_message function."""
        # async def on_message(bt, message)
        bt, message = mock.Mock(), mock.Mock()
        await bot.LGBot.on_message(bt, message)
        ptch.assert_called_once_with(bt, message)

    @mock.patch("lgrez.bot._on_raw_reaction_add")
    async def test_on_raw_reaction_add(self, ptch):
        """Unit tests for bot.on_raw_reaction_add function."""
        # async def on_raw_reaction_add(bt, payload)
        bt, payload = mock.Mock(), mock.Mock()
        await bot.LGBot.on_raw_reaction_add(bt, payload)
        ptch.assert_called_once_with(bt, payload)

    @mock.patch("lgrez.bot._on_command_error")
    async def test_on_command_error(self, ptch):
        """Unit tests for bot.on_command_error function."""
        # async def on_command_error(bt, ctx, exc)
        bt, ctx, exc = mock.Mock(), mock.Mock(), mock.Mock()
        await bot.LGBot.on_command_error(bt, ctx, exc)
        ptch.assert_called_once_with(bt, ctx, exc)

    @mock.patch("lgrez.bot._on_error")
    async def test_on_error(self, ptch):
        """Unit tests for bot.on_error function."""
        # async def on_error(bt, event, *args, **kwargs)
        bt, event = mock.Mock(), mock.Mock()
        args = [mock.Mock(), mock.Mock(), mock.Mock()]
        kwargs = {"a": mock.Mock(), "bzzt": mock.Mock()}
        await bot.LGBot.on_error(bt, event, *args, **kwargs)
        ptch.assert_called_once_with(bt, event, *args, **kwargs)


    async def test_i_am_alive(self):
        """Unit tests for bot.i_am_alive function."""
        # def i_am_alive(self, filename="alive.log")
        i_am_alive = bot.LGBot.i_am_alive
        bt = mock.Mock()

        # default filename
        with freezegun.freeze_time(datetime.datetime.utcfromtimestamp(75824)):
            with mock.patch("builtins.open", mock.mock_open()) as open_mock:
                i_am_alive(bt)
        open_mock.assert_called_once_with("alive.log", "w")
        open_mock.return_value.write.assert_called_once_with("75824.0")
        bt.loop.call_later.assert_called_once_with(60, bt.i_am_alive,
                                                   "alive.log")
        open_mock.reset_mock()
        bt.reset_mock()

        # custom filename
        with freezegun.freeze_time(datetime.datetime.utcfromtimestamp(75824)):
            with mock.patch("builtins.open", mock.mock_open()) as open_mock:
                i_am_alive(bt, "krkrkr")
        open_mock.assert_called_once_with("krkrkr", "w")
        open_mock.return_value.write.assert_called_once_with("75824.0")
        bt.loop.call_later.assert_called_once_with(60, bt.i_am_alive, "krkrkr")


    @mock_env.patch_env(LGREZ_DISCORD_TOKEN="t@ken",
                        LGREZ_SERVER_ID="12345")
    @mock.patch("lgrez.bdd.connect")
    @mock.patch("builtins.print")
    async def test_run(self, ppatch, connect_patch):
        """Unit tests for bot.run function."""
        # def run(self, **kwargs)
        run = bot.LGBot.run
        bt = mock.Mock(bot.LGBot)
        kwargs = {"a": mock.Mock(), "bzzt": mock.Mock()}

        # default filename
        with mock.patch("discord.ext.commands.Bot.run") as run_patch:
            run(bt, **kwargs)
        self.assertEqual(ppatch.call_count, 5)
        vers, codb, codbok, codi, deco = ppatch.call_args_list
        self.assertIn(__version__, vers.args[0])
        self.assertEqual(bt.GUILD_ID, 12345)
        self.assertIn("Connecting", codb.args[0])
        connect_patch.assert_called_once()
        self.assertIn("Connected", codbok.args[0])
        self.assertEqual(config.bot, bt)
        del config.bot
        self.assertIn("Connecting", codi.args[0])
        run_patch.assert_called_once_with("t@ken", **kwargs)
        self.assertIn("Disconnected", deco.args[0])



"""lg-rez / features / IA des réponses

Tout ce qui concerne la manière dont le bot réagit aux messages :
détermination de la meilleure réaction, gestion des réactions,
activation/désactivation des modes de chat

"""

import re
import random
import requests

from discord.ext import commands

from lgrez import config
from lgrez.blocs import tools
from lgrez.bdd import Trigger, Reaction, Role


# Marqueurs de séparation du mini-langage des séquences-réactions
MARK_OR = ' <||> '
MARK_THEN = ' <&&> '
MARK_REACT = '<::>'
MARK_CMD = '<!!>'
MARKS = [MARK_OR, MARK_THEN, MARK_REACT, MARK_CMD]


async def _build_sequence(ctx):
    """Construction d'une séquence-réaction par l'utilisateur"""
    reponse = ""
    fini = False
    while not fini:
        message = await ctx.send(
            "Réaction du bot : prochain message/commande/média, "
            "ou réaction à ce message"
        )
        ret = await tools.wait_for_react_clic(
            message, process_text=True, trigger_all_reacts=True,
            trigger_on_commands=True
        )
        if isinstance(ret, str):
            if ret.startswith(config.bot.command_prefix):   # Commande
                reponse += MARK_CMD + ret.lstrip(config.bot.command_prefix)
            else:                                           # Texte / média
                reponse += ret
        else:                                               # React
            reponse += MARK_REACT + ret.name

        message = await ctx.send("▶ Puis / 🔀 Ou / ⏹ Fin ?")
        ret = await tools.wait_for_react_clic(
            message, emojis={"▶": MARK_THEN, "🔀": MARK_OR, "⏹": False}
        )
        if ret:
            # On ajoute la marque OR ou THEN à la séquence
            reponse += ret
        else:
            fini = True

    return reponse


def fetch_tenor(trigger):
    """Renvoie le GIF Tenor le plus "pertinent" pour un texte donné.

    Args:
        trigger (str): texte auquel réagir.

    Returns:
        :class:`str` (URL du GIF) ou ``None``
    """
    # API key module ternorpy (parce que la flemme de créer un compte Tenor)
    apikey = "J5UVWPVIM4A5"

    rep = requests.get(
        url="https://api.tenor.com/v1/search",
        params={
            "q": trigger, "key": apikey, "limit": 1,
            "locale": "fr_FR", "contentfilter": "off",
            "media_filter": "minimal", "ar_range": "all"
        }
    )

    if rep:
        gifs = rep.json()["results"]
        # Payload Tenor : {..., "results":[ ... ]}
        # (https://tenor.com/gifapi/documentation#responseobjects-gif)
        if gifs:
            return gifs[0]["itemurl"]

    return None     # Pas de GIF trouvé


class GestionIA(commands.Cog):
    """Commandes relatives à l'IA (réponses automatiques du bot)"""

    @commands.command()
    @tools.private
    async def stfu(self, ctx, force=None):
        """Active/désactive la réponse automatique du bot sur ton channel privé

        Args:
            force: ``"start"``/``"on"`` / ``"stop"``/``"off"`` permet de
                forcer l'activation / la désactivation.

        Sans argument, la commande active les réactions si désactivées
        et vice-versa ; avec un autre argument, elle le fait silencieusment.

        N'agit que sur les messages classiques envoyés dans le channel :
        les commandes restent reconnues.

        Si vous ne comprenez pas le nom de la commande, demandez à Google.
        """
        id = ctx.channel.id

        if force in [None, "start", "on"] and id not in config.bot.in_stfu:
            config.bot.in_stfu.append(id)
            await ctx.send(
                "Okay, je me tais ! Tape !stfu quand tu voudras de "
                "nouveau de moi :cry:"
            )

        elif force in [None, "stop", "off"] and id in config.bot.in_stfu:
            config.bot.in_stfu.remove(id)
            await ctx.send("Ahhh, ça fait plaisir de pouvoir reparler !")

        elif force not in ["start", "on", "stop", "off"]:
            # Quelque chose d'autre que start/stop précisé après !stfu :
            # bot discret
            if id in config.bot.in_stfu:
                config.bot.in_stfu.remove(id)
            else:
                config.bot.in_stfu.append(id)


    @commands.command(aliases=["cancer", "214"])
    async def fals(self, ctx, force=None):
        """Active/désactive le mode « foire à la saucisse »

        Args:
            force: ``"start"``/``"on"`` / ``"stop"``/``"off"`` permet de
                forcer l'activation / la désactivation.

        Sans argument, la commande active le mode si désactivé
        et vice-versa.

        En mode « foire à la saucisse », le bot réagira à (presque) tous
        les messages, pas seulement sur les motifs qu'on lui a appris.

        À utiliser à vos risques et périls !
        """
        id = ctx.channel.id

        if force in [None, "start", "on"] and id not in config.bot.in_fals:
            config.bot.in_fals.append(id)
            await ctx.send(
                "https://tenor.com/view/saucisse-sausage-gif-5426973"
            )

        elif force in [None, "stop", "off"] and id in config.bot.in_fals:
            config.bot.in_fals.remove(id)
            await ctx.send("T'as raison, faut pas abuser des bonnes choses")

        elif force not in ["start", "on", "stop", "off"]:
            # Quelque chose d'autre que start/stop précisé après !fals :
            # bot discret
            if id in config.bot.in_fals:
                config.bot.in_fals.remove(id)
            else:
                config.bot.in_fals.append(id)


    @commands.command(aliases=["r"])
    async def react(self, ctx, *, trigger):
        """Force le bot à réagir à un message

        Args:
            trigger: texte auquel le bot doit réagir

        Permet de faire appel à l'IA du bot même sur les chans publics,
        ou en mode STFU, etc.

        Si utilisée par un MJ, active aussi le mode débug des
        évaluations Python (messages d'erreur).
        """
        oc = ctx.message.content
        ctx.message.content = trigger
        debug = (ctx.message.webhook_id
                 or ctx.author.top_role == config.Role.mj)
        await process_IA(ctx.message, debug=debug)
        ctx.message.content = oc
        # On rétablit le message original pour ne pas qu'il trigger
        # l'IA 2 fois, le cas échéant


    @commands.command(aliases=["rf"])
    async def reactfals(self, ctx, *, trigger):
        """Force le bot à réagir à un message comme en mode Foire à la saucisse

        Args:
            trigger: texte auquel le bot doit réagir

        Permet de faire appel directement au mode Foire à la saucisse,
        même si il n'est pas activé / sur un chan public.
        """
        async with ctx.typing():
            gif = fetch_tenor(trigger)

        if gif:
            await ctx.send(gif)
        else:
            await ctx.send("Palaref")


    @commands.command()
    @tools.mjs_et_redacteurs
    async def addIA(self, ctx, *, triggers=None):
        """Ajoute une règle d'IA (COMMANDE MJ/RÉDACTEURS)

        Args:
            triggers: mot(s), phrase(s), ou expression(s) séparées par
                des points-virgules ou sauts de lignes (ne peut pas
                contenir de ">" ou ";" hors séparateurs)

        Nouveau système d'ajout rapide :
            !addIA trigger > reponse
            !addIA trig1 ; trig2 > reponse

        ou, si réponse à un message de contenu "trigger" :
            !addIA > reponse        (triggers = [trigger])
            !addIA trig2 > reponse  (triggers = [trigger, trig2])

        Une sécurité empêche d'ajouter un trigger déjà existant.

        Dans le cas où plusieurs expressions sont spécifiées, toutes
        déclencheront l'action demandée.
        """
        fast = False
        if not triggers:
            await ctx.send(
                "Mots/expressions déclencheurs (non sensibles à la "
                "casse / accents), séparés par des points-virgules "
                "ou des sauts de ligne :"
            )
            mess = await tools.wait_for_message_here(ctx)
            triggers = mess.content

        elif ">" in triggers:
            fast = True
            # Mode rapide
            triggers, reponse = triggers.split(">", maxsplit=1)

        triggers = triggers.replace('\n', ';').split(';')
        if fast and ctx.message.reference:
            repmess = ctx.message.reference.resolved
            if repmess:
                triggers.append(repmess.content)

        triggers = [tools.remove_accents(s).lower().strip() for s in triggers]
        triggers = list({trig for trig in triggers if trig})
        # filtre doublons (accents et non accents...) et triggers vides

        for trigger in triggers.copy():
            if Trigger.query.filter_by(trigger=trigger).all():
                await ctx.send(f"Trigger `{trigger}` déjà associé à une "
                               "réaction, enlevé")
                triggers.remove(trigger)

        if not triggers:
            await ctx.send("Aucun trigger valide, abort")
            return

        await ctx.send(f"Triggers : `{'` – `'.join(triggers)}`")

        if fast:
            reponse = reponse.strip()
        else:
            reponse = await _build_sequence(ctx)

        await ctx.send(f"Résumé de la séquence : {tools.code(reponse)}")
        async with ctx.typing():
            reac = Reaction(reponse=reponse)
            config.session.add(reac)

            trigs = [Trigger(trigger=trigger, reaction=reac)
                     for trigger in triggers]
            config.session.add_all(trigs)

            config.session.commit()

        await ctx.send("Règle ajoutée en base.")


    @commands.command()
    @tools.mjs_et_redacteurs
    async def listIA(self, ctx, trigger=None, sensi=0.5):
        """Liste les règles d'IA reconnues par le bot (COMMANDE MJ/RÉDACTEURS)

        Args
            trigger (optionnel): mot/expression permettant de filter et
                trier les résultats. SI ``trigger`` FAIT PLUS D'UN MOT,
                IL DOIT ÊTRE ENTOURÉ PAR DES GUILLEMETS !
            sensi: sensibilité de détection (ratio des caractères
                correspondants, entre 0 et 1) si trigger est précisé.
        """
        async with ctx.typing():
            if trigger:
                trigs = Trigger.find_nearest(
                    trigger, col=Trigger.trigger,
                    sensi=sensi, solo_si_parfait=False
                )
                if not trigs:
                    await ctx.send(
                        f"Rien trouvé, pas de chance (sensi = {sensi})"
                    )
                    return
            else:
                raw_trigs = Trigger.query.order_by(Trigger.id).all()
                # Trié par date de création
                trigs = list(zip(raw_trigs, [None] * len(raw_trigs)))
                # Mise au format (trig, score)

            reacts = []     # Réactions associées à notre liste de triggers
            for trig in trigs:
                if (reac := trig[0].reaction) not in reacts:
                    # Pas de doublons, et reste ordonné
                    reacts.append(reac)

            def nettoy(s):
                # Abrège la réponse si trop longue et neutralise les
                # sauts de ligne / rupture code_bloc, pour affichage
                s = s.replace('\r\n', '\\n').replace('\n', '\\n')
                s = s.replace('\r', '\\r').replace("```", "'''")
                if len(s) < 75:
                    return s
                else:
                    return s[:50] + " [...] " + s[-15:]

            rep = ""
            for reac in reacts:                 # pour chaque réponse
                r = ""
                for (trig, score) in trigs:         # pour chaque trigger
                    if trig.reaction == reac:
                        sc = f"({float(score):.2}) " if score else ""
                        r += f" - {sc}{trig.trigger}"
                        # (score) trigger - (score) trigger ...

                rep += r.ljust(50) + f" ⇒ {nettoy(reac.reponse)}\n"
                # ⇒ réponse

            rep += "\nPour modifier une réaction, utiliser !modifIA <trigger>."

        await tools.send_code_blocs(ctx, rep)
        # On envoie, en séparant en blocs de 2000 caractères max


    @commands.command()
    @tools.mjs_et_redacteurs
    async def modifIA(self, ctx, *, trigger=None):
        """Modifie/supprime une règle d'IA (COMMANDE MJ/RÉDACTEURS)

        Args:
            trigger: mot/expression déclenchant la réaction à
                modifier/supprimer

        Permet d'ajouter et supprimer des triggers, de modifier la
        réaction du bot (construction d'une séquence de réponses
        successives ou aléatoires) ou de supprimer la réaction.
        """
        if not trigger:
            await ctx.send(
                "Mot/expression déclencheur de la réaction à modifier :"
            )
            mess = await tools.wait_for_message_here(ctx)
            trigger = mess.content

        trigs = Trigger.find_nearest(trigger, col=Trigger.trigger)
        if not trigs:
            await ctx.send("Rien trouvé.")
            return

        trig = trigs[0][0]
        reac = trig.reaction

        displ_seq = (reac.reponse if reac.reponse.startswith('`')
                     else tools.code(reac.reponse))     # Pour affichage
        trigs = list(reac.triggers)

        await ctx.send(
            f"Triggers : `{'` – `'.join([trig.trigger for trig in trigs])}`\n"
            f"Séquence réponse : {displ_seq}"
        )

        message = await ctx.send(
            "Modifier : ⏩ triggers / ⏺ Réponse / 🚮 Supprimer ?"
        )
        choix = await tools.wait_for_react_clic(
            message, emojis={"⏩": 1, "⏺": 2, "🚮": 0})

        if choix == 1:              # Modification des triggers
            fini = False
            while not fini:
                s = "Supprimer un trigger : \n"
                for i, t in enumerate(trigs[:10]):
                    s += f"{tools.emoji_chiffre(i+1)}. {t.trigger} \n"

                mess = await ctx.send(
                    s + "Ou entrer un mot / une expression pour l'ajouter "
                    "en trigger.\n⏹ pour finir"
                )
                r = await tools.wait_for_react_clic(
                    mess,
                    emojis={(tools.emoji_chiffre(i) if i else "⏹"): str(i)
                            for i in range(len(trigs) + 1)},
                    process_text=True
                )

                if r == "0":
                    fini = True
                elif r.isdigit() and (n := int(r)) <= len(trigs):
                    config.session.delete(trigs[n - 1])
                    config.session.commit()
                    del trigs[n - 1]
                else:
                    new_trig = Trigger(trigger=r, reaction=reac)
                    trigs.append(new_trig)
                    config.session.add(trig)
                    config.session.commit()

            if not trigs:        # on a tout supprimé !
                await ctx.send(
                    "Tous les triggers supprimés, suppression de la réaction"
                )
                config.session.delete(reac)
                config.session.commit()
                return

        elif choix == 2:            # Modification de la réponse
            if any([mark in reac.reponse for mark in MARKS]):
                # Séquence compliquée
                await ctx.send(
                    "\nLa séquence-réponse peut être refaite manuellement "
                    "ou modifiée rapidement en envoyant directment la "
                    "séquence ci-dessus modifiée (avec les marqueurs : "
                    f"OU = {tools.code(MARK_OR)}, "
                    f"ET = {tools.code(MARK_THEN)}, "
                    f"REACT = {tools.code(MARK_REACT)}, "
                    f"CMD = {tools.code(MARK_CMD)})"
                )

            reponse = await _build_sequence(ctx)
            reac.reponse = reponse

        else:                       # Suppression
            config.session.delete(reac)
            for trig in trigs:
                config.session.delete(trig)

        config.session.commit()

        await ctx.send("Fini.")


async def trigger_at_mj(message):
    """Règle d'IA : réaction si le message mentionne les MJs.

    Args:
        message (~discord.Message): message auquel réagir.

    Returns:
        - ``True`` -- si le message mentionne les MJ et qu'une réponse
          a été envoyée
        - ``False`` -- sinon
    """
    if config.Role.mj in message.role_mentions:
        await message.channel.send(
            "Les MJs ont entenu ton appel, ils sont en route ! :superhero:"
        )
        return True

    return False


async def trigger_roles(message, sensi=0.8):
    """Règle d'IA : réaction si un nom de rôle est donné.

    Args:
        message (~discord.Message): message auquel réagir.
        sensi (float): sensibilité de la recherche
            (voir :meth:`.bdd.base.TableMeta.find_nearest`).

    Trouve l'entrée la plus proche de ``message.content`` dans la table
    :class:`.bdd.Role`.

    Returns:
        - ``True`` -- si un rôle a été trouvé (sensibilité ``> sensi``)
          et qu'une réponse a été envoyée
        - ``False`` -- sinon
    """
    roles = Role.find_nearest(message.content, col=Role.nom,
                              filtre=(Role.actif.is_(True)), sensi=sensi)

    if roles:       # Au moins un trigger trouvé à cette sensi
        await message.channel.send(embed=roles[0][0].embed)
        return True

    return False


async def trigger_reactions(message, chain=None, sensi=0.7, debug=False):
    """Règle d'IA : réaction à partir de la table :class:`.bdd.Reaction`.

    Args:
        message (~discord.Message`): message auquel réagir.
        chain (str): contenu auquel réagir (défaut : contenu de
            ``message``).
        sensi (float): sensibilité de la recherche (cf
            :meth:`.bdd.base.TableMeta.find_nearest`).
        debug (bool): si ``True``, affiche les erreurs lors de
            l'évaluation des messages (voir :func:`.tools.eval_accols`).

    Trouve l'entrée la plus proche de ``chain`` dans la table
    :class:`.bdd.Reaction` ; si il contient des accolades, évalue le
    message selon le contexte de ``message``.

    Returns:
        - ``True`` -- si une réaction a été trouvé (sensibilité
          ``> sensi``) et qu'une réponse a été envoyée
        - ``False`` -- sinon
    """
    if not chain:                   # Si pas précisé,
        chain = message.content         # contenu de message
    trigs = Trigger.find_nearest(chain, col=Trigger.trigger, sensi=sensi)

    if trigs:       # Au moins un trigger trouvé à cette sensi
        trig = trigs[0][0]                  # Meilleur trigger (score max)
        seq = trig.reaction.reponse         # Séquence-réponse associée

        for rep in seq.split(MARK_THEN):        # Pour chaque étape :
            if MARK_OR in rep:
                # Si plusieurs possiblités, on en choisit une random
                rep = random.choice(rep.split(MARK_OR))

            if rep.startswith(MARK_REACT):          # Réaction
                react = rep.lstrip(MARK_REACT)
                emoji = tools.emoji(react, must_be_found=False) or react
                await message.add_reaction(emoji)

            elif rep.startswith(MARK_CMD):          # Commande
                message.content = rep.replace(MARK_CMD,
                                              config.bot.command_prefix)
                # Exécution de la commande
                await config.bot.process_commands(message)

            else:                                   # Sinon, texte / média
                # On remplace tous les "{expr}" par leur évaluation
                rep = tools.eval_accols(rep, locals_=locals(), debug=debug)
                await message.channel.send(rep)

        return True

    return False


async def trigger_sub_reactions(message, sensi=0.9, debug=False):
    """Règle d'IA : réaction à partir de la table, mais sur les mots

    Appelle :func:`trigger_reactions(bot, message, mot, sensi, debug)
    <.trigger_reactions>` pour tous les mots ``mot`` composant
    ``message.content`` (mots de plus de 4 lettres, testés des plus
    longs aux plus courts).

    Args:
        message (~discord.Message`): message auquel réagir.
        sensi (float): sensibilité de la recherche (cf
            :meth:`.bdd.base.TableMeta.find_nearest`).
        debug (bool): si ``True``, affiche les erreurs lors de
            l'évaluation des messages (voir :func:`.tools.eval_accols`).

    Returns:
        - ``True`` -- si une réaction a été trouvé (sensibilité
          ``> sensi``) et qu'une réponse a été envoyée
        - ``False`` -- sinon
    """
    mots = message.content.split(" ")
    if len(mots) > 1:       # Si le message fait plus d'un mot
        for mot in sorted(mots, key=lambda m: -len(m)):
            # On parcourt les mots du plus long au plus court
            if len(mot) > 4:        # on élimine les mots de liaison
                if await trigger_reactions(message, chain=mot,
                                           sensi=sensi, debug=debug):
                    # Si on trouve une sous-rect (à 0.9)
                    return True

    return False


async def trigger_di(message):
    """Règle d'IA : réaction aux messages en di... / cri...

    Args:
        message (~discord.Message): message auquel réagir.

    Returns:
        - ``True`` -- si le message correspond et qu'une réponse a été
          envoyée
        - ``False`` -- sinon
    """
    c = message.content
    diprefs = ["di", "dy", "dis ", "dit ", "dis-", "dit-"]
    criprefs = ["cri", "cry", "kri", "kry"]
    pos_prefs = {c.lower().find(pref): pref for pref in diprefs + criprefs
                 if pref in c[:-1].lower()}
    # On extrait les cas où le préfixe est à la fin du message

    if pos_prefs:           # Si on a trouvé au moins un préfixe
        i = min(pos_prefs)
        pref = pos_prefs[i]
        if pref in criprefs:
            mess = tools.bold(c[i + len(pref):].upper())
        else:
            mess = c[i + len(pref):]
        await message.channel.send(mess, tts=True)
        # On envoie le di.../cri... en mode TTS (oh si, c'est rigolo)
        return True

    return False


async def trigger_gif(message):
    """Règle d'IA : réaction par GIF en mode Foire à la saucisse.

    Args:
        message (~discord.Message): message auquel réagir

    Returns:
        - ``True`` -- si le message correspond et qu'une réponse a été
          envoyée
        - ``False`` -- sinon
    """
    if message.channel.id in config.bot.in_fals:
        # Chan en mode Foire à la saucisse
        async with message.channel.typing():
            gif = fetch_tenor(message.content)
            if gif:
                await message.channel.send(gif)
                return True

    return False


async def trigger_mot_unique(message):
    """Règle d'IA : réaction à un mot unique (le répète).

    Args:
        message (~discord.Message): message auquel réagir.

    Returns:
        - ``True`` -- si le message correspond et qu'une réponse a été
          envoyée
        - ``False`` -- sinon
    """
    if len(message.content.split()) == 1 and ":" not in message.content:
        # : pour ne pas trigger aux liens
        rep = f"{message.content.capitalize()} ?"
        await message.channel.send(rep)
        return True

    return False


async def trigger_a_ou_b(message):
    """Règle d'IA : réaction à un motif type « a ou b » (répond « b »).

    Args:
        message (~discord.Message): message auquel réagir.

    Returns:
        - ``True`` -- si le message correspond et qu'une réponse a été
          envoyée
        - ``False`` -- sinon
    """
    if (motif := re.fullmatch(r"(.+)\s+ou\s+(.+?)", message.content)):
        rep = f"{motif.group(2).rstrip(' !?.,;')}.".capitalize()
        await message.channel.send(rep)
        return True

    return False


async def default(message):
    """Règle d'IA : réponse par défaut

    Args:
        message (~discord.Message): message auquel réagir.

    Returns:
        - ``True`` -- si le message correspond et qu'une réponse a été
          envoyée
        - ``False`` -- sinon
    """
    mess = "Désolé, je n'ai pas compris :person_shrugging:"
    if random.random() < 0.05:
        mess += "\n(et toi, tu as perdu)"
    await message.channel.send(mess)
    return True


async def process_IA(message, debug=False):
    """Exécute les règles d'IA.

    Args:
        message (~discord.Message): message auquel réagir.
        debug (bool): si ``True``, affiche les erreurs lors de
            l'évaluation des messages (voir :func:`.tools.eval_accols`).
    """
    (
        await trigger_at_mj(message)            # @MJ (aled)
        or await trigger_gif(message)           # Un petit GIF ? (si FALS)
        or await trigger_roles(message)         # Rôles
        or await trigger_reactions(             # Table Reaction ("IA")
            message, debug=debug)
        or await trigger_sub_reactions(         # IA sur les mots
            message, debug=debug)
        or await trigger_a_ou_b(message)        # "a ou b" ==> "b"
        or await trigger_di(message)            # di... / cri...
        or await trigger_mot_unique(message)    # Un seul mot ==> on répète
        or await default(message)               # Réponse par défaut
    )



"""lg-rez / LGBot

Classe principale

"""

import logging
import sys
import time
import traceback

import discord
from discord.ext import commands

from lgrez import __version__, config, bdd
from lgrez.blocs import env, tools, one_command, ready_check
from lgrez.features import *        # Tous les sous-modules


#: str: Description par défaut du bot
default_descr = "LG-bot – Plateforme pour parties endiablées de Loup-Garou"


async def _check_and_prepare_objects(bot):
    errors = []

    def prepare_attributes(rc_class, discord_type, converter):
        """Rend prêt les attributs d'une classe ReadyCheck"""
        for attr in rc_class:
            raw = rc_class.get_raw(attr)
            # Si déjà prêt, on actualise quand même (reconnexion)
            name = raw.name if isinstance(raw, discord_type) else raw
            try:
                ready = converter(name)
            except ValueError:
                qualname = f"config.{rc_class.__name__}.{attr}"
                errors.append(f"{discord_type.__name__} {qualname} = "
                              f"\"{name}\" non trouvé !")
            else:
                setattr(rc_class, attr, ready)

    prepare_attributes(config.Role, discord.Role, tools.role)
    prepare_attributes(config.Channel, discord.TextChannel, tools.channel)
    prepare_attributes(config.Emoji, discord.Emoji, tools.emoji)

    try:
        tools.channel(config.private_chan_category_name)
    except ValueError:
        errors.append(f"catégorie config.private_chan_category_name = "
                      f"\"{config.private_chan_category_name}\" non trouvée")

    try:
        tools.channel(config.boudoirs_category_name)
    except ValueError:
        errors.append(f"catégorie config.boudoirs_category_name = "
                      f"\"{config.boudoirs_category_name}\" non trouvée")

    if len(errors) > config._missing_objects:
        # Nouvelles erreurs
        msg = (f"LGBot.on_ready: {len(errors)} errors:\n - "
               + "\n - ".join(errors))
        logging.error(msg)

        try:
            atmj = config.Role.mj.mention
        except ready_check.NotReadyError:
            atmj = "@everyone"

        try:
            await tools.log(msg, code=True, prefixe=f"{atmj} ERREURS :")
        except ready_check.NotReadyError:
            config.Channel.logs = config.guild.text_channels[0]
            msg += "\n-- Routing logs to this channel."
            await tools.log(msg, code=True, prefixe=f"{atmj} ERREURS :")

    elif len(errors) < config._missing_objects:
        if errors:
            # Erreurs résolues, il en reste
            msg = f"{len(errors)} errors:\n - " + "\n - ".join(errors)
            logging.error(msg)
            await tools.log(msg, code=True, prefixe=f"Erreurs restantes :")
        else:
            # Toutes erreurs résolues
            await tools.log("Configuration rôles/chans/emojis OK.")

    config._missing_objects = len(errors)

    # Webhook
    existing = await config.Channel.logs.webhooks()

    if existing:
        config.webhook = existing[0]
    else:           # Création du webhook
        config.webhook = await config.Channel.logs.create_webhook(
            name=bot.user.name,
            avatar=await bot.user.avatar_url.read()
        )
        await tools.log(f"Webhook de tâches planifiées créé")


# ---- Réactions aux différents évènements

# Au démarrage du bot
async def _on_ready(bot):
    config.loop = bot.loop          # Enregistrement loop

    guild = bot.get_guild(bot.GUILD_ID)
    if not guild:
        raise RuntimeError(f"on_ready : Serveur d'ID {bot.GUILD_ID} "
                           "(``LGREZ_SERVER_ID``) introuvable")

    print(f"      Connected to '{guild.name}'! "
          f"({len(guild.channels)} channels, {len(guild.members)} members)")

    if config.output_liveness:
        bot.i_am_alive()            # Start liveness regular output

    print("[3/3] Initialization (bot.on_ready)...")

    # Préparations des objects globaux
    config.guild = guild
    await bot.check_and_prepare_objects()

    await tools.log("Just rebooted!")
    await bot.change_presence(activity=discord.Activity(
        type=discord.ActivityType.listening,
        name="vos demandes (!help)"
    ))

    # Tâches planifiées
    taches = bdd.Tache.query.all()
    for tache in taches:
        # Si action manquée, l'exécute immédiatement, sinon l'enregistre
        tache.register()

    if taches:
        await tools.log(f"{len(taches)} tâches planifiées récupérées "
                        "en base et reprogrammées.")

    print("      Initialization complete.")
    print("\nListening for events.")


# À l'arrivée d'un membre sur le serveur
async def _on_member_join(bot, member):
    if member.guild != config.guild:        # Mauvais serveur
        return

    await tools.log(f"Arrivée de {member.name}#{member.discriminator} "
                    "sur le serveur")
    await inscription.main(member)


# Au départ d'un membre du serveur
async def _on_member_remove(bot, member):
    if member.guild != config.guild:        # Mauvais serveur
        return

    await tools.log(
        f"{tools.mention_MJ(member)} ALERTE : départ du serveur de "
        f"{member.display_name} ({member.name}#{member.discriminator}) !")


# À chaque message
async def _on_message(bot, message):
    if message.author == bot.user:          # Pas de boucles infinies
        return

    if not message.guild:                   # Message privé
        await message.channel.send(
            "Je n'accepte pas les messages privés, désolé !"
        )
        return

    if message.guild != config.guild:       # Mauvais serveur
        return

    if (not message.webhook_id              # Pas un webhook
        and message.author.top_role == config.Role.everyone):
        # Pas de rôle affecté : le bot te calcule même pas
        return

    if message.content.startswith(bot.command_prefix + " "):
        message.content = bot.command_prefix + message.content[2:]

    # On trigger toutes les commandes
    # (ne PAS remplacer par bot.process_commands(message), en théorie
    # c'est la même chose mais ça détecte pas les webhooks...)
    ctx = await bot.get_context(message)
    await bot.invoke(ctx)

    if (not message.content.startswith(bot.command_prefix)
        and message.channel.name.startswith(config.private_chan_prefix)
        and message.channel.id not in bot.in_command
        and message.channel.id not in bot.in_stfu):
        # Conditions d'IA respectées (voir doc) : on trigger
        await IA.process_IA(message)


# À chaque réaction ajoutée
async def _on_raw_reaction_add(bot, payload):
    reactor = payload.member
    if reactor == bot.user:                         # Boucle infinie
        return

    if payload.guild_id != config.guild.id:         # Mauvais serveur
        return

    chan = config.guild.get_channel(payload.channel_id)
    if not chan.name.startswith(config.private_chan_prefix):
        # Pas dans un chan privé
        return

    if config.Role.joueur_en_vie not in reactor.roles:
        # Pas un joueur en vie
        return

    if payload.emoji == config.Emoji.bucher:
        ctx = await tools.create_context(reactor, "!vote")
        await ctx.send(
            f"{payload.emoji} > "
            + tools.bold("Vote pour le condamné du jour :")
        )
        await bot.invoke(ctx)       # On trigger !vote

    elif payload.emoji == config.Emoji.maire:
        ctx = await tools.create_context(reactor, "!votemaire")
        await ctx.send(
            f"{payload.emoji} > "
            + tools.bold("Vote pour le nouveau maire :")
        )
        await bot.invoke(ctx)       # On trigger !votemaire

    elif payload.emoji == config.Emoji.lune:
        ctx = await tools.create_context(reactor, "!voteloups")
        await ctx.send(
            f"{payload.emoji} > "
            + tools.bold("Vote pour la victime des loups :")
        )
        await bot.invoke(ctx)       # On trigger !voteloups

    elif payload.emoji == config.Emoji.action:
        ctx = await tools.create_context(reactor, "!action")
        await ctx.send(f"{payload.emoji} > " + tools.bold("Action :"))
        await bot.invoke(ctx)       # On trigger !voteloups


# ---- Gestion des erreurs

def _showexc(exc):
    return f"{type(exc).__name__}: {exc}"


# Gestion des erreurs dans les commandes
async def _on_command_error(bot, ctx, exc):
    if ctx.guild != config.guild:               # Mauvais serveur
        return

    if isinstance(exc, commands.CommandInvokeError):
        if isinstance(exc.original, tools.CommandExit):     # STOP envoyé
            await ctx.send(str(exc.original) or "Mission aborted.")
            return

        if isinstance(exc.original,                         # Erreur BDD
                      (bdd.SQLAlchemyError, bdd.DriverOperationalError)):
            try:
                config.session.rollback()           # On rollback la session
                await tools.log("Rollback session")
            except ready_check.NotReadyError:
                pass

        # Dans tous les cas (sauf STOP), si erreur à l'exécution
        prefixe = ("Oups ! Un problème est survenu à l'exécution de "
                   "la commande  :grimacing: :")

        if ctx.message.webhook_id or ctx.author.top_role >= config.Role.mj:
            # MJ / webhook : affiche le traceback complet
            e = traceback.format_exception(type(exc.original), exc.original,
                                           exc.original.__traceback__)
            await tools.send_code_blocs(ctx, "".join(e), prefixe=prefixe)
        else:
            # Pas MJ : exception seulement
            await ctx.send(f"{prefixe}\n{tools.mention_MJ(ctx)} ALED – "
                           + tools.ital(_showexc(exc.original)))

    elif isinstance(exc, commands.CommandNotFound):
        await ctx.send(
            f"Hum, je ne connais pas cette commande  :thinking:\n"
            f"Utilise {tools.code('!help')} pour voir la liste des commandes."
        )

    elif isinstance(exc, commands.DisabledCommand):
        await ctx.send("Cette commande est désactivée. Pas de chance !")

    elif isinstance(exc, (commands.ConversionError, commands.UserInputError)):
        await ctx.send(
            f"Hmm, ce n'est pas comme ça qu'on utilise cette commande ! "
            f"({tools.code(_showexc(exc))})\n*Tape "
            f"`!help {ctx.invoked_with}` pour plus d'informations.*"
        )
        # ctx.message.content = f"!help {ctx.command.name}"
        # ctx = await bot.get_context(ctx.message)
        # await ctx.reinvoke()

    elif isinstance(exc, commands.CheckAnyFailure):
        # Normalement raise que par @tools.mjs_only
        await ctx.send(
            "Hé ho toi, cette commande est réservée aux MJs !  :angry:"
        )

    elif isinstance(exc, commands.MissingAnyRole):
        # Normalement raise que par @tools.joueurs_only
        await ctx.send(
            "Cette commande est réservée aux joueurs ! "
            "(parce qu'ils doivent être inscrits en base, toussa) "
            f"({tools.code('!doas')} est là en cas de besoin)"
        )

    elif isinstance(exc, commands.MissingRole):
        # Normalement raise que par @tools.vivants_only
        await ctx.send(
            "Désolé, cette commande est réservée aux joueurs en vie !"
        )

    elif isinstance(exc, one_command.AlreadyInCommand):
        if ctx.command.name in ["addIA", "modifIA"]:
            # addIA / modifIA : droit d'enregistrer les commandes, donc chut
            return
        await ctx.send(
            f"Impossible d'utiliser une commande pendant "
            "un processus ! (vote...)\n"
            f"Envoie {tools.code(config.stop_keywords[0])} "
            "pour arrêter le processus."
        )

    elif isinstance(exc, commands.CheckFailure):
        # Autre check non vérifié
        await ctx.send(
            f"Tiens, il semblerait que cette commande ne puisse "
            f"pas être exécutée ! {tools.mention_MJ(ctx)} ?\n"
            f"({tools.ital(_showexc(exc))})")

    else:
        await ctx.send(
            f"Oups ! Une erreur inattendue est survenue  :grimacing:\n"
            f"{tools.mention_MJ(ctx)} ALED – {tools.ital(_showexc(exc))}"
        )


# Erreurs non gérées par le code précédent (hors cadre d'une commande)
async def _on_error(bot, event, *args, **kwargs):
    etype, exc, tb = sys.exc_info()     # Exception ayant causé l'appel

    if isinstance(exc, (bdd.SQLAlchemyError,            # Erreur SQL
                        bdd.DriverOperationalError)):
        try:
            config.session.rollback()       # On rollback la session
            await tools.log("Rollback session")
        except ready_check.NotReadyError:
            pass

    await tools.log(
        traceback.format_exc(),
        code=True,
        prefixe=f"{config.Role.mj.mention} ALED : Exception Python !"
    )

    # On remonte l'exception à Python (pour log, ne casse pas la loop)
    raise


# ---- Définition classe principale

class LGBot(commands.Bot):
    """Bot Discord pour parties de Loup-Garou à la PCéenne.

    Classe fille de :class:`discord.ext.commands.Bot`, implémentant les
    commandes et fonctionnalités du Loup-Garou de la Rez.

    Args:
        command_prefix (str): passé à :class:`discord.ext.commands.Bot`
        case_insensitive (bool): passé à
            :class:`discord.ext.commands.Bot`
        description (str): idem, défaut \:
            :attr:`lgrez.bot.default_descr`
        intents (discord.Intents): idem, défaut \:
            :meth:`~discord.Intents.all()`. *Certaines commandes et
            fonctionnalités risquent de ne pas fonctionner avec une
            autre valeur.*
        member_cache_flags (discord.MemberCacheFlags): idem, défaut \:
            :meth:`~discord.MemberCacheFlags.all()`. *Certaines
            commandes et fonctionnalités risquent de ne pas fonctionner
            avec une autre valeur.*
        \*\*kwargs: autres options de :class:`~discord.ext.commands.Bot`

    Warning:
        LG-Bot n'est **pas** thread-safe : seule une instance du bot
        peut tourner en parallèle dans un interpréteur.

        (Ceci est du aux objets de :mod:`.config`, contenant directement
        le bot, le serveur Discord, la session de connexion BDD... ;
        cette limitation résulte d'une orientation volontaire du module
        depuis sa version 2.0 pour simplifier et optimiser la
        manipulation des objects et fonctions).

    Attributes:
        bot (int): L'ID du serveur sur lequel tourne le bot (normalement
            toujours :attr:`config.guild` ``.id``).  Vaut ``None`` avant
            l'appel à :meth:`run`, puis la valeur de la variable
            d'environnement ``LGREZ_SERVER_ID``.
        in_command (list[int]): IDs des salons dans lequels une
            commande est en cours d'exécution.
        in_stfu (list[int]): IDs des salons en mode STFU.
        in_fals (list[int]): IDs des salons en mode Foire à la saucisse.
        tasks (dict[int (.bdd.Tache.id), asyncio.TimerHandle]):
            Tâches planifiées actuellement en attente. Privilégier
            plutôt l'emploi de :attr:`.bdd.Tache.handler`.

    """
    def __init__(self, command_prefix="!", case_insensitive=True,
                 description=None, intents=None, member_cache_flags=None,
                 **kwargs):
        """Initialize self"""
        # Paramètres par défaut
        if description is None:
            description = default_descr
        if intents is None:
            intents = discord.Intents.all()
        if member_cache_flags is None:
            member_cache_flags = discord.MemberCacheFlags.all()

        # Construction du bot Discord.py
        super().__init__(
            command_prefix=command_prefix,
            description=description,
            case_insensitive=case_insensitive,
            intents=intents,
            member_cache_flags=member_cache_flags,
            **kwargs
        )

        # Définition attribus personnalisés
        self.GUILD_ID = None
        self.in_stfu = []
        self.in_fals = []
        self.tasks = {}

        # Système de limitation à une commande à la fois
        self.in_command = []
        self.add_check(one_command.not_in_command)
        self.before_invoke(one_command.add_to_in_command)
        self.after_invoke(one_command.remove_from_in_command)

        # Commandes joueur : information, actions privés et publiques
        self.add_cog(informations.Informations(self))
        self.add_cog(voter_agir.VoterAgir(self))
        self.add_cog(actions_publiques.ActionsPubliques(self))
        # Commandes MJs : gestion votes/actions, synchro GSheets,
        # planifications, posts et embeds...
        self.add_cog(open_close.OpenClose(self))
        self.add_cog(sync.Sync(self))
        self.add_cog(taches.GestionTaches(self))
        self.add_cog(communication.Communication(self))
        # Commandes mixtes : comportement de l'IA et trucs divers
        self.add_cog(IA.GestionIA(self))
        self.add_cog(annexe.Annexe(self))
        self.add_cog(chans.GestionChans(self))
        # Commandes spéciales, méta-commandes...
        self.remove_command("help")
        self.add_cog(special.Special(self))

    # Réactions aux différents évènements
    async def on_ready(self):
        """Méthode appellée par Discord au démarrage du bot.

        Vérifie le serveur (appelle :meth:`check_and_prepare_objects`),
        log et affiche publiquement que le bot est fonctionnel
        (activité) ;  restaure les tâches planifiées éventuelles
        et exécute celles manquées.

        Si :attr:`config.output_liveness` vaut ``True``, lance
        :attr:`bot.i_am_alive <.LGBot.i_am_alive>`
        (écriture chaque minute sur un fichier disque)

        Voir :func:`discord.on_ready` pour plus d'informations.
        """
        await _on_ready(self)

    async def on_member_join(self, member):
        """Méthode appellée par l'API à l'arrivée d'un nouveau membre.

        Log et lance le processus d'inscription.

        Ne fait rien si l'arrivée n'est pas sur le serveur
        :attr:`config.guild`.

        Args:
            member (discord.Member): Le membre qui vient d'arriver.

        Voir :func:`discord.on_member_join` pour plus d'informations.
        """
        await _on_member_join(self, member)

    async def on_member_remove(self, member):
        """Méthode appellée par l'API au départ d'un membre du serveur.

        Log en mentionnant les MJs.

        Ne fait rien si le départ n'est pas du serveur
        :attr:`config.guild`.

        Args:
            member (discord.Member): Le joueur qui vient de partir.

        Voir :func:`discord.on_member_remove` pour plus d'informations.
        """
        await _on_member_remove(self, member)

    async def on_message(self, message):
        """Méthode appellée par l'API à la réception d'un message.

        Invoque l'ensemble des commandes, ou les règles d'IA si
            - Le message n'est pas une commande
            - Le message est posté dans un channel privé (dont le nom
              commence par :attr:`config.private_chan_prefix`)
            - Il n'y a pas déjà de commande en cours dans ce channel
            - Le channel n'est pas en mode STFU

        Ne fait rien si le message n'est pas sur le serveur
        :attr:`config.guild`, si il est envoyé par le bot lui-même
        ou par un membre sans aucun rôle affecté.

        Args:
            member (discord.Member): Le joueur qui vient d'arriver.

        Voir :func:`discord.on_message` pour plus d'informations.
        """
        await _on_message(self, message)

    async def on_raw_reaction_add(self, payload):
        """Méthode appellée par l'API à l'ajout d'une réaction.

        Appelle la fonction adéquate si le membre est un joueur
        inscrit, est sur un chan de conversation bot et a cliqué sur
        :attr:`config.Emoji.bucher`, :attr:`~config.Emoji.maire`,
        :attr:`~config.Emoji.lune` ou :attr:`~config.Emoji.action`.

        Ne fait rien si la réaction n'est pas sur le serveur
        :attr:`config.guild`.

        Args:
            payload (discord.RawReactionActionEvent): Paramètre
                limité (car le message n'est pas forcément dans le
                cache du bot, par exemple si il a été reboot depuis).

        Quelques attributs utiles :
          - ``payload.member`` (:class:`discord.Member`) : Membre
            ayant posé la réaction
          - ``payload.emoji`` (:class:`discord.PartialEmoji`) :
            PartialEmoji envoyé
          - ``payload.message_id`` (:class:`int`) : ID du message réacté

        Voir :func:`discord.on_raw_reaction_add` pour plus
        d'informations.
        """
        await _on_raw_reaction_add(self, payload)

    # Gestion des erreurs
    async def on_command_error(self, ctx, exc):
        """Méthode appellée par l'API à un exception dans une commande.

        Analyse l'erreur survenue et informe le joueur de manière
        adéquate en fonction, en mentionnant les MJs si besoin.

        Ne fait rien si l'exception n'a pas eu lieu sur le serveur
        :attr:`config.guild`.

        Args:
            ctx (discord.ext.commands.Context): Contexte dans lequel
                l'exception a été levée
            exc (discord.ext.commands.CommandError): Exception levée

        Voir :func:`discord.on_command_error` pour plus d'informations.
        """
        await _on_command_error(self, ctx, exc)

    async def on_error(self, event, *args, **kwargs):
        """Méthode appellée par l'API à une exception hors commande.

        Log en mentionnant les MJs. Cette méthode permet de gérer les
        exceptions sans briser la loop du bot (i.e. il reste en ligne).

        Args:
            event (str): Nom de l'évènement ayant généré une erreur
                (``"member_join"``, ``"message"``...)
            *args, \**kwargs: Arguments passés à la fonction traitant
                l'évènement : ``member``, ``message``...

        Voir :func:`discord.on_error` pour plus d'informations.
        """
        await _on_error(self, event, *args, **kwargs)

    # Checks en temps réels des modifs des objets nécessaires au bot
    async def check_and_prepare_objects(self):
        """Vérifie et prépare les objets Discord nécessaires au bot.

        Remplit :class:`.config.Role`, :class:`.config.Channel`,
        :class:`.config.Emoji`, :attr:`config.private_chan_category_name`,
        :attr:`config.boudoirs_category_name` et :attr:`config.webhook`
        avec les objets Discord correspondants, et avertit les MJs en cas
        d'éléments manquants.
        """
        await _check_and_prepare_objects(self)

    async def on_guild_channel_delete(self, channel):
        if channel.guild == config.guild:
            await self.check_and_prepare_objects()

    async def on_guild_channel_update(self, before, after):
        if before.guild == config.guild and config._missing_objects:
            await self.check_and_prepare_objects()

    async def on_guild_channel_create(self, channel):
        if channel.guild == config.guild and config._missing_objects:
            await self.check_and_prepare_objects()

    async def on_guild_role_delete(self, role):
        if role.guild == config.guild:
            await self.check_and_prepare_objects()

    async def on_guild_role_update(self, before, after):
        if before.guild == config.guild and config._missing_objects:
            await self.check_and_prepare_objects()

    async def on_guild_role_create(self, channel):
        if role.guild == config.guild and config._missing_objects:
            await self.check_and_prepare_objects()

    async def on_guild_emojis_update(self, guild, before, after):
        if guild == config.guild:
            await self.check_and_prepare_objects()

    async def on_webhooks_update(self, channel):
        if channel == config.Channel.logs:
            await self.check_and_prepare_objects()

    # Système de vérification de vie
    def i_am_alive(self, filename="alive.log"):
        """Témoigne que le bot est en vie et non bloqué.

        Exporte le temps actuel (UTC) et planifie un nouvel appel
        dans 60s. Ce processus n'est lancé que si
        :attr:`config.output_liveness` est mis à ``True``.

        Args:
            filename (:class:`str`): fichier où exporter le temps
                actuel (écrase le contenu).
        """
        with open(filename, "w") as f:
            f.write(str(time.time()))
        self.loop.call_later(60, self.i_am_alive, filename)

    # Lancement du bot
    def run(self, **kwargs):
        """Prépare puis lance le bot (bloquant).

        Récupère les informations de connexion, établit la connexion
        à la base de données puis lance le bot.

        Args:
            \**kwargs: Passés à :meth:`discord.ext.commands.Bot.run`.
        """
        print(f"--- LGBot v{__version__} ---")

        # Récupération du token du bot et de l'ID du serveur
        LGREZ_DISCORD_TOKEN = env.load("LGREZ_DISCORD_TOKEN")
        self.GUILD_ID = int(env.load("LGREZ_SERVER_ID"))

        # Connection BDD
        print("[1/3] Connecting to database...")
        bdd.connect()
        print("      Connected!")

        # Enregistrement
        config.bot = self

        # Lancement du bot (bloquant)
        print("[2/3] Connecting to Discord...")
        super().run(LGREZ_DISCORD_TOKEN, **kwargs)

        print("\nDisconnected.")



import unittest
from unittest import mock

from lgrez.blocs import env



class TestEnv(unittest.IsolatedAsyncioTestCase):
    """Unit tests for lgrez.blocs.env module."""

    @mock.patch("os.getenv")
    def test_load(self, ge_patch):
        """Unit tests for env.load function."""
        # def load(VAR_NAME)
        load = env.load
        # existing
        ge_patch.return_value = "12345"
        VAR_NAME = "OUI"
        result = load(VAR_NAME)
        ge_patch.assert_called_once_with("OUI")
        ge_patch.reset_mock()
        self.assertEqual(result, "12345")
        # non existing
        ge_patch.return_value = None
        VAR_NAME = "NON"
        with self.assertRaises(RuntimeError):
            load(VAR_NAME)
        ge_patch.assert_called_once_with("NON")



import unittest
from unittest import mock

from lgrez.bdd import model_ia



class TestReaction(unittest.IsolatedAsyncioTestCase):
    """Unit tests for lgrez.bdd.model_ia.Reaction methods."""

    def test___repr__(self):
        """Unit tests for Reaction.__repr__ method."""
        # def __repr__(self)
        __repr__ = model_ia.Reaction.__repr__
        # < 15
        slf = mock.Mock(model_ia.Reaction, id=11, reponse="blebelebel")
        rpr = __repr__(slf)
        self.assertEqual(rpr, "<Reaction #11 (blebelebel...)>")
        # > 15
        slf = mock.Mock(model_ia.Reaction, id=11,
                        reponse="blebelebelpof^psdmlsqd*^ùfmqsfpspdofsdf")
        rpr = __repr__(slf)
        self.assertEqual(rpr, "<Reaction #11 (blebelebelpof^p...)>")


class TestTrigger(unittest.IsolatedAsyncioTestCase):
    """Unit tests for lgrez.bdd.model_ia.Trigger methods."""

    def test___repr__(self):
        """Unit tests for Trigger.__repr__ method."""
        # def __repr__(self)
        __repr__ = model_ia.Trigger.__repr__
        # < 15
        slf = mock.Mock(model_ia.Trigger, id=11, trigger="blebelebel")
        rpr = __repr__(slf)
        self.assertEqual(rpr, "<Trigger #11 (blebelebel...)>")
        # > 15
        slf = mock.Mock(model_ia.Trigger, id=11,
                        trigger="blebelebelpof^psdmlsqd*^ùfmqsfpspdofsdf")
        rpr = __repr__(slf)
        self.assertEqual(rpr, "<Trigger #11 (blebelebelpof^p...)>")
